#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import textwrap
from typing import Union
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
STD_LENGTH = 79  # Standard length for output to console
# ==================================================================================================================================


class Colors:
    # Define colors used throughout this framework
    #
    # Attributes:
    #     WARN    (str): Defines color for warnings.
    #     END     (str): Defines end of color in string.
    BANNERA = '\033[93m'
    BANNERB = '\033[94m'
    WARN = '\033[91m'
    END = '\033[0m'


def header(program: str, version: str, commit: str, length: int = STD_LENGTH) -> None:
    """ Print big header with program name and logo to console.

        Args:
            length (int): Number of characters used within each line.
    """
    # string = 'Parametric Exploration and Control Engine'
    print(Colors.BANNERA + '┏' + '━'*(length-1))
    # print(Colors.BANNERA + '┃')
    print(Colors.BANNERA + '┃' + ' P y H O P E — Python High-Order Preprocessing Environment')
    # print(Colors.BANNERA + '┃' + ' {}'.format(string))
    print(Colors.BANNERA + '┃' + Colors.END + ' {} version {} [commit {}]'.format(program, version, commit))
    print(Colors.BANNERA + '┡' + '━'*(length-1) + Colors.END)


def banner(string: str, length: int = STD_LENGTH) -> None:
    """ Print the input `string` in a banner-like output.

        Args:
            string (str): String to be printed in banner.
            length (int): (Optional.) Number of characters in each line.
    """
    print(Colors.BANNERA + '\n' + '='*length)
    print(Colors.BANNERA + ' '+string)
    print(Colors.BANNERA + '='*length + Colors.END)


def small_banner(string: str, length: int = STD_LENGTH) -> None:
    """ Print the input `string` in a small banner-like output.

        Args:
            string (str): String to be printed in banner.
            length (int): (Optional.) Number of characters in each line.
    """
    print(Colors.BANNERB + '\n' + '-'*length)
    print(Colors.BANNERB + ' '+string)
    print(Colors.BANNERB + '-'*length + Colors.END)


def warn(string: str, length: int = STD_LENGTH) -> str:
    """ Format the input `string` as a warning with the corresponding color.

        Args:
                string (str): String to be printed in banner.
                length (int): (Optional.) Number of characters in each line.
    """
    prefix   = Colors.WARN + ' WARNING  ┃ '  + Colors.END
    lprefix  = len(' WARNING  ┃ ')
    wrap_msg = textwrap.fill(string, width=length - lprefix)

    # Add prefix to each line
    format_msg = '\n'.join(f'{prefix}{line}' for line in wrap_msg.splitlines())
    return format_msg


def warning(string: str) -> None:
    """ Print the input `string` as a warning with the corresponding color.

        Args:
            string (str): String to be printed in banner.
            length (int): (Optional.) Number of characters in each line.
    """
    print(Colors.WARN + '\n !! '+string+' !! \n' + Colors.END)


def sep(length: int = 5) -> None:
    print('├' + '─'*(length-1))


def separator(length: int = 46) -> None:
    print('├' + '─'*(length-1))


def end(program: str, time: float, length: int = STD_LENGTH) -> None:
    print('┢' + '━'*(length-1))
    print('┃ {} completed in [{:.2f} sec]'.format(program, time))
    print('┗' + '━'*(length-1))


def info(string: str, newline: bool = False, end: Union[str, None] = None) -> None:
    """ Print the input `string` as generic output without special formatting.

        Args:
            string (str): String to be printed in banner.
            length (int): (Optional.) Number of characters in each line.
    """
    if newline:
        print('\n│ '+ string, end=end)
    else:
        print('│ '  + string, end=end)


def routine(string: str, newline=False) -> None:
    """ Print the input `string` as generic output without special formatting.

        Args:
            string (str): String to be printed in banner.
            length (int): (Optional.) Number of characters in each line.
    """
    if newline:
        print('\n├── ' + string)
    else:
        print('├── '   + string)


def printoption(option: str, value: str, status: str, length: int = 31) -> None:
    """ Print the input `string` as option string

        Args:
            string (str): String to be printed in banner.
            length (int): (Optional.) Number of characters in each line.
    """
    try:
        if len(value) > length:
            pvalue = '{}...'.format(value[:(length-3)])
        else:
            pvalue = value
    except TypeError:
        pvalue = value
    print(f'│ {option:>{length}} │ {pvalue:<{length}} │ {status} │')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import copy
import sys
import traceback
from typing import Union
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
from scipy import spatial
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.output.output as hopout
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def flip_analytic(side: list, nbside: list) -> int:
    """ Determines the flip of the side-to-side connection based on the analytic side ID
        flip = 1 : 1st node of neighbor side = 1st node of side
        flip = 2 : 2nd node of neighbor side = 1st node of side
        flip = 3 : 3rd node of neighbor side = 1st node of side
        flip = 4 : 4th node of neighbor side = 1st node of side
    """
    # Local imports ----------------------------------------
    from pyhope.common.common import find_index
    # ------------------------------------------------------
    return find_index(nbside, side[0])


def flip_physical(corners: np.ndarray, nbcorners: np.ndarray, tol: float, msg: str) -> int:
    """ Determines the flip of the side-to-side connection based on the physical positions
        flip = 1 : 1st node of neighbor side = 1st node of side
        flip = 2 : 2nd node of neighbor side = 1st node of side
        flip = 3 : 3rd node of neighbor side = 1st node of side
        flip = 4 : 4th node of neighbor side = 1st node of side
    """
    ptree     = spatial.KDTree(nbcorners)

    trCorn    = ptree.query(corners[0])
    if trCorn[0] > tol:
        hopout.warning(f'Could not determine flip of {msg} side within tolerance {tol}, exiting...')
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)

    flipID    = trCorn[1] + 1
    return flipID


def connect_sides(sideIDs: list, sides: list, flipID: int) -> None:
    """ Connect the master and slave sides
    """
    sides[sideIDs[0]].update({
        # Master side contains positive global side ID
        'MS'        : 1,
        'Connection': sideIDs[1],
        'Flip'      : 0,
        'nbLocSide' : sides[sideIDs[1]]['LocSide']
    })
    sides[sideIDs[1]].update({
        'MS'        : 0,
        'Connection': sideIDs[0],
        'Flip'      : flipID,
        'nbLocSide' : sides[sideIDs[0]]['LocSide']
    })


def find_bc_index(bcs: list, key: str) -> Union[int, None]:
    """ Find the index of a BC from its name in the list of BCs
    """
    for iBC, bc in enumerate(bcs):
        if key in bc['Name']:
            return iBC
        # Try again without the leading 'BC_'
        if key[:3] == 'BC_' and key[3:] in bc['Name']:
            return iBC
    return None


def find_closest_side(points: np.ndarray, stree: spatial._kdtree.cKDTree, tol: float, msg: str) -> int:
    """ Query the tree for the closest side
    """
    trSide = stree.query(points)

    # Check if the found side is within tolerance
    # trSide contains the Euclidean distance and the index of the
    # opposing side in the nbFaceSet
    if trSide[0] > tol:
        hopout.warning(f'Could not find {msg} side within tolerance {tol}, exiting...')
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)
    return trSide[1]


def get_side_id(corners: np.ndarray, side_dict: dict) -> int:
    """ Get sorted corners and hash them to get the side ID
    """
    corners_sorted = np.sort(corners)
    corners_hash = hash(corners_sorted.tobytes())
    return side_dict[corners_hash][0]


def ConnectMesh() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.common.common import find_index
    from pyhope.io.io_vars import MeshFormat
    # ------------------------------------------------------

    match io_vars.outputformat:
        case MeshFormat.FORMAT_HDF5:
            pass
        case _:
            return

    hopout.separator()
    hopout.info('CONNECT MESH...')

    mesh    = mesh_vars.mesh
    elems   = mesh_vars.elems
    sides   = mesh_vars.sides

    # cell_sets contain the face IDs [dim=2]
    # > Offset is calculated with entities from [dim=0, dim=1]
    offsetcs = 0
    for key, value in mesh.cells_dict.items():
        if 'vertex' in key:
            offsetcs += value.shape[0]
        elif 'line' in key:
            offsetcs += value.shape[0]

    # Map sides to BC
    # > Create a dict containing only the face corners
    side_corners = dict()
    # for iSide, side in enumerate(sides):
    #     corners = np.sort(side['Corners'])
    #     corners = hash(corners.tobytes())
    #     side_corners.update({iSide: corners})
    for iElem, elem in enumerate(elems):
        for iSide, side in enumerate(elem['Sides']):
            corners = np.sort(sides[side]['Corners'])
            corners = hash(corners.tobytes())
            side_corners.update({side: corners})

    # Build the reverse dictionary
    corner_side = dict()
    for key, val in side_corners.items():
        if val not in corner_side:
            corner_side[val] = [key]
        else:
            corner_side[val].append(key)

    # Try to connect the inner sides
    ninner = 0
    for index, (key, val) in enumerate(corner_side.items()):
        match len(val):
            case 1:  # BC side
                continue
            case 2:  # Internal side
                sideIDs   = val
                corners   = sides[sideIDs[0]]['Corners']
                nbcorners = sides[sideIDs[1]]['Corners']
                flipID    = flip_analytic(corners, nbcorners) + 1
                # Connect the sides
                connect_sides(sideIDs, sides, flipID)
                ninner += 1
            case _:  # Zero or more than 2 sides
                hopout.warning('Found internal side with more than two adjacent elements, exiting...')
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)

    # Set BC and periodic sides
    bcs = mesh_vars.bcs
    vvs = mesh_vars.vvs
    csetMap = []
    for key, cset in mesh.cell_sets.items():
        # Check if the set is a BC
        bcID = find_bc_index(bcs, key)
        if bcID is None:
            hopout.warning(f'Could not find BC {key} in list, exiting...')
            sys.exit(1)

        # Find the mapping to the (N-1)-dim elements
        csetMap = [s for s in range(len(cset)) if cset[s] is not None]

        # Get the list of sides
        for iMap in csetMap:
            iBCsides = np.array(cset[iMap]).astype(int) - offsetcs
            mapFaces = mesh.cells[iMap].data

            # Map the unique quad sides to our non-unique elem sides
            for iSide in iBCsides:
                # Get the quad corner nodes
                # FIXME: HARDCODED FIRST 4 NODES WHICH ARE THE OUTER CORNER NODES FOR QUADS!
                corners = np.sort(np.array(mapFaces[iSide][0:4]))
                corners = hash(corners.tobytes())

                # Boundary faces are unique
                # sideID  = find_key(face_corners, corners)
                sideID = corner_side[corners][0]
                sides[sideID].update({'BCID': bcID})

    # Try to connect the periodic sides
    # TODO: SET ANOTHER TOLERANCE
    # tol = 5.#5.E-1

    for key, cset in mesh.cell_sets.items():
        # Check if the set is a BC
        bcID = find_bc_index(bcs, key)
        if bcID is None:
            hopout.warning(f'Could not find BC {key} in list, exiting...')
            sys.exit(1)

        # Only periodic BCs and only try to connect in positive direction
        if bcs[bcID]['Type'][0] != 1 or bcs[bcID]['Type'][3] < 0:
            continue

        # Get the opposite side
        iVV        = bcs[bcID]['Type'][3] - 1
        nbType     = copy.copy(bcs[bcID]['Type'])
        nbType[3] *= -1
        nbBCID     = find_index([s['Type'] for s in bcs], nbType)
        nbBCName   = bcs[nbBCID]['Name']

        # Collapse all opposing corner nodes into an [:, 12] array
        nbCellSet  = mesh.cell_sets[nbBCName]
        # Find the mapping to the (N-1)-dim elements
        nbcsetMap  = [s for s in range(len(nbCellSet)) if nbCellSet[s] is not None and nbCellSet[s].size > 0]

        # FIXME: TODO HYBRID MESHES
        if len(nbcsetMap) > 1:
            print('Hybrid meshes currently not supported')
            sys.exit(1)

        # Get the list of sides
        nbFaceSet  =  np.array(nbCellSet[csetMap[0]]).astype(int)
        nbmapFaces = mesh.cells[csetMap[0]].data
        nbCorners  = [np.array(nbmapFaces[s - offsetcs]) for s in nbFaceSet]
        nbPoints   = copy.copy(np.sort(mesh.points[nbCorners], axis=1))
        nbPoints   = nbPoints.reshape(nbPoints.shape[0], nbPoints.shape[1]*nbPoints.shape[2])
        del nbCorners

        # Build a k-dimensional tree of all points on the opposing side
        stree    = spatial.KDTree(nbPoints)
        # Get the list of sides on our side
        iBCsides = np.array(cset[csetMap[0]]).astype(int) - offsetcs

        # Map the unique quad sides to our non-unique elem sides
        for iSide in iBCsides:
            # Get the quad corner nodes
            corners   = np.array(nbmapFaces[iSide])
            points    = copy.copy(mesh.points[corners])

            # Shift the points in periodic direction
            for iPoint in range(points.shape[0]):
                points[iPoint, :] += vvs[iVV]['Dir']
            points    = np.sort(points, axis=0).flatten()

            # Query the try for the opposing side
            tol       = np.linalg.norm(vvs[iVV]['Dir'], ord=2).astype(float) * 1.E-1
            nbSideIdx = find_closest_side(points, stree, tol, 'periodic')
            nbiSide   = nbFaceSet[nbSideIdx] - offsetcs

            # Get our and neighbor corner quad nodes
            sideID    = get_side_id(nbmapFaces[iSide  ][0:4], corner_side)
            nbSideID  = get_side_id(nbmapFaces[nbiSide][0:4], corner_side)

            # Build the connection, including flip
            sideIDs   = [sideID, nbSideID]
            points    = mesh.points[sides[sideIDs[0]]['Corners']]
            for iPoint in range(points.shape[0]):
                points[iPoint, :] += vvs[iVV]['Dir']

            # > Find the first neighbor point to determine the flip
            nbcorners = mesh.points[sides[sideIDs[1]]['Corners']]
            flipID    = flip_physical(points, nbcorners, tol, 'periodic')

            # Connect the sides
            connect_sides(sideIDs, sides, flipID)

    # Non-connected sides without BCID are possible inner sides
    nConnSide = [s for s in sides if 'Connection' not in s and 'BCID' not in s]
    # Append the inner BCs
    for s in (s for s in sides if 'BCID' in s and 'Connection' not in s):
        if mesh_vars.bcs[s['BCID']]['Type'][0] == 0:
            nConnSide.append(s)

    nInterZoneConnect = len(nConnSide)

    hopout.separator()

    # Loop over all sides and try to connect
    while len(nConnSide) > 1:
        # Remove the first side from the list
        targetSide = nConnSide.pop(0)

        # Collapse all opposing corner nodes into an [:, 12] array
        nbCorners  = [s['Corners'] for s in nConnSide]
        nbPoints   = copy.copy(np.sort(mesh.points[nbCorners], axis=1))
        nbPoints   = nbPoints.reshape(nbPoints.shape[0], nbPoints.shape[1]*nbPoints.shape[2])
        del nbCorners

        # Build a k-dimensional tree of all points on the opposing side
        stree      = spatial.KDTree(nbPoints)

        # Map the unique quad sides to our non-unique elem sides
        corners    = targetSide['Corners']
        points     = np.sort(mesh.points[corners], axis=0).flatten()

        # Query the tree for the opposing side
        tol = 1.E-10
        nbSideIdx = find_closest_side(points, stree, tol, 'internal')
        nbiSide   = nbSideIdx

        # Get our and neighbor corner quad nodes
        sideID    = get_side_id(targetSide[        'Corners'], corner_side)
        nbSideID  = get_side_id(nConnSide[nbiSide]['Corners'], corner_side)

        # Build the connection, including flip
        sideIDs   = [sideID, nbSideID]
        points    = mesh.points[sides[sideIDs[0]]['Corners']]
        # > Find the first neighbor point to determine the flip
        nbcorners = mesh.points[sides[sideIDs[1]]['Corners']]
        flipID    = flip_physical(points, nbcorners, tol, 'periodic')

        # Connect the sides
        connect_sides(sideIDs, sides, flipID)

        # Update the list
        nConnSide = [s for s in sides if 'Connection' not in s and 'BCID' not in s]
        # Append the inner BCs
        for s in (s for s in sides if 'BCID' in s and 'Connection' not in s):
            if mesh_vars.bcs[s['BCID']]['Type'][0] == 0:
                nConnSide.append(s)

    if nInterZoneConnect > 0:
        hopout.sep()
        hopout.routine('Connected {} inter-zone faces'.format(nInterZoneConnect))

    # Set the global side ID
    globalSideID = 0
    for iSide, side in enumerate(sides):
        # Already counted the side
        if 'GlobalSideID' in side:
            continue

        globalSideID += 1
        if 'Connection' not in side:  # BC side
            side.update({'GlobalSideID':  globalSideID })
        elif side['MS'] == 1:         # Internal / periodic side (master side)
            # Master side does not have a flip
            # Set the positive globalSideID of the master side
            side.update({'GlobalSideID':  globalSideID })
            # Set the negative globalSideID of the slave  side
            nbSideID = side['Connection']
            sides[nbSideID].update({'GlobalSideID': -(globalSideID)})

    # Count the sides
    nsides         = len(sides)
    sides_conn     = np.array(['Connection' in side for side in sides])  # noqa: E272
    sides_bc       = np.array(['BCID'       in side for side in sides])  # noqa: E272

    # Count each type of side
    ninnersides    = np.sum( sides_conn & ~sides_bc)
    nperiodicsides = np.sum( sides_conn &  sides_bc)
    nbcsides       = np.sum(~sides_conn &  sides_bc)

    hopout.sep()
    hopout.info(' Number of sides          : {:12d}'.format(nsides))
    hopout.info(' Number of inner sides    : {:12d}'.format(ninnersides))
    hopout.info(' Number of boundary sides : {:12d}'.format(nbcsides))
    hopout.info(' Number of periodic sides : {:12d}'.format(nperiodicsides))
    hopout.sep()

    hopout.info('CONNECT MESH DONE!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def faces() -> list:
    """ Return a list of all sides of a hexahedron
    """
    return ['z-', 'y-', 'x+', 'y+', 'x-', 'z+']


def edge_to_dir(edge: int) -> int:
    """ GMSH: Create edges from points in the given direction
    """
    match edge:
        case 0 | 2 |  4 |  6:
            return 0
        case 1 | 3 |  5 |  7:
            return 1
        case 8 | 9 | 10 | 11:
            return 2
        case _:
            print('Error in edge_to_dir, unknown edge')
            sys.exit(1)


def face_to_edge(face, dtype=int) -> np.ndarray:
    """ GMSH: Create faces from edges in the given direction
    """
    match face:
        case 'z-':
            return np.array([  0,  1,   2,   3], dtype=dtype)
        case 'y-':
            return np.array([  0,  9,  -4,  -8], dtype=dtype)
        case 'x+':
            return np.array([  1, 10,  -5,  -9], dtype=dtype)
        case 'y+':
            return np.array([ -2, 10,   6, -11], dtype=dtype)
        case 'x-':
            return np.array([  8, -7, -11,   3], dtype=dtype)
        case 'z+':
            return np.array([  4,  5,   6,   7], dtype=dtype)
        case _:  # Default
            print('Error in face_to_edge, unknown face')
            sys.exit(1)


def face_to_corner(face, dtype=int) -> np.ndarray:
    """ GMSH: Get points on faces in the given direction
    """
    match face:
        case 'z-':
            return np.array([  0,  1,   2,   3], dtype=dtype)
        case 'y-':
            return np.array([  0,  1,   5,   4], dtype=dtype)
        case 'x+':
            return np.array([  1,  2,   6,   5], dtype=dtype)
        case 'y+':
            return np.array([  2,  6,   7,   3], dtype=dtype)
        case 'x-':
            return np.array([  0,  4,   7,   3], dtype=dtype)
        case 'z+':
            return np.array([  4,  5,   6,   7], dtype=dtype)
        case _:  # Default
            print('Error in face_to_corner, unknown face')
            sys.exit(1)


def face_to_cgns(face, dtype=int) -> np.ndarray:
    """ CGNS: Get points on faces in the given direction
    """
    match face:
        case 'z-':
            return np.array([  0,  3,   2,   1], dtype=dtype)
        case 'y-':
            return np.array([  0,  1,   5,   4], dtype=dtype)
        case 'x+':
            return np.array([  1,  2,   6,   5], dtype=dtype)
        case 'y+':
            return np.array([  2,  3,   7,   6], dtype=dtype)
        case 'x-':
            return np.array([  0,  4,   7,   3], dtype=dtype)
        case 'z+':
            return np.array([  4,  5,   6,   7], dtype=dtype)
        case _:  # Default
            print('Error in face_to_corner, unknown face')
            sys.exit(1)


def count_elems(mesh: meshio._mesh.Mesh) -> int:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    nElems = 0
    for iType, elemType in enumerate(mesh.cells_dict.keys()):
        # Only consider three-dimensional types
        if not any(s in elemType for s in mesh_vars.ELEM.type.keys()):
            continue

        ioelems = mesh.get_cells_type(elemType)
        nElems += ioelems.shape[0]
    return nElems


def centeroidnp(coords: np.ndarray) -> np.ndarray:
    """ Compute the centroid (barycenter) of a set of coordinates
    """
    length = coords.shape[0]
    sum_x  = np.sum(coords[:, 0])
    sum_y  = np.sum(coords[:, 1])
    sum_z  = np.sum(coords[:, 2])
    return np.array([sum_x/length, sum_y/length, sum_z/length])


def calc_elem_bary(mesh: meshio._mesh.Mesh) -> list:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    nElems   = count_elems(mesh)
    elemBary = [np.ndarray(3)] * nElems
    for iType, elemType in enumerate(mesh.cells_dict.keys()):
        # Only consider three-dimensional types
        if not any(s in elemType for s in mesh_vars.ELEM.type.keys()):
            continue

        ioelems = mesh.get_cells_type(elemType)

        for elemID, cell in enumerate(ioelems):
            elemBary[elemID] = centeroidnp(mesh_vars.mesh.points[cell])
    return elemBary



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
import traceback
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def DefineMesh() -> None:
    """ Define general options for mesh generation / readin
    """
    # Local imports ----------------------------------------
    from pyhope.readintools.readintools import CreateInt, CreateIntArray, CreateRealArray, CreateSection, CreateStr, CreateLogical
    # ------------------------------------------------------

    CreateSection('Mesh')
    CreateInt(      'Mode',                               help='Mesh generation mode (1 - Internal, 2 - External [MeshIO])')
    # Internal mesh generator
    CreateInt(      'nZones',                             help='Number of mesh zones')
    CreateRealArray('Corner',         24,  multiple=True, help='Corner node positions: (/ x_1,y_1,z_1,, x_2,y_2,z_2,, '
                                                                                        '... ,, x_8,y_8,z_8/)')  # noqa: E127
    CreateIntArray( 'nElems',          3,  multiple=True, help='Number of elements in each direction')
    CreateStr(      'BoundaryName',        multiple=True, help='Name of domain boundary')
    CreateIntArray( 'BoundaryType',    4,  multiple=True, help='(/ Type, curveIndex, State, alpha /)')
    CreateIntArray( 'BCIndex',         6,  multiple=True, help='Index of BC for each boundary face')
    CreateRealArray('vv',              3,  multiple=True, help='Vector for periodic BC')
    # External mesh readin through GMSH
    CreateStr(      'Filename',            multiple=True, help='Name of external mesh file')
    CreateLogical(  'MeshIsAlreadyCurved', default=False, help='Enables mesh agglomeration')
    # Common settings
    CreateInt(      'NGeo'         ,       default=1,     help='Order of spline-reconstruction for curved surfaces')
    CreateInt(      'BoundaryOrder',       default=2,     help='Order of spline-reconstruction for curved surfaces (legacy)')
    CreateLogical(  'doSortIJK',           default=False, help='Sort the mesh elements along the I,J,K directions')
    CreateLogical(  'CheckElemJacobians',  default=True,  help='Check the Jacobian and scaled Jacobian for each element')


def InitMesh() -> None:
    """ Readin general option for mesh generation / readin
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import GetInt, CountOption
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('INIT MESH...')

    mesh_vars.mode = GetInt('Mode')

    NGeo     = GetInt('NGeo')          if CountOption('NGeo')          else None
    BCOrder  = GetInt('BoundaryOrder') if CountOption('BoundaryOrder') else None

    if not NGeo and not BCOrder:
        mesh_vars.nGeo = 1
    elif NGeo and BCOrder and NGeo != BCOrder - 1:
        hopout.warning('NGeo / BoundaryOrder must be equal to NGeo + 1!')
        sys.exit(1)
    else:
        mesh_vars.nGeo = NGeo or (BCOrder - 1)

        if mesh_vars.nGeo < 1:
            hopout.warning('Effective boundary order < 1. Try increasing the NGeo / BoundaryOrder parameter!')
            sys.exit(1)

    hopout.info('INIT MESH DONE!')


def GenerateMesh() -> None:
    """ Generate the mesh
        Mode 1 - Use internal mesh generator
        Mode 2 - Readin external mesh through GMSH
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.mesh.mesh_builtin import MeshCartesian
    from pyhope.mesh.mesh_external import MeshExternal
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('GENERATE MESH...')

    match mesh_vars.mode:
        case 1:  # Internal Cartesian Mesh
            mesh = MeshCartesian()
        case 3:  # External mesh
            mesh = MeshExternal()
        case _:  # Default
            hopout.warning('Unknown mesh mode {}, exiting...'.format(mesh_vars.mode))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)

    mesh_vars.mesh = mesh

    hopout.info('GENERATE MESH DONE!')


def RegenerateMesh() -> None:
    """ Finish missing mesh information such as BCs
    """
    # Local imports ----------------------------------------
    from pyhope.mesh.mesh_external import BCCGNS
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    match mesh_vars.mode:
        case 1:  # Internal Cartesian Mesh
            mesh = mesh_vars.mesh
        case 3:  # External CGNS mesh
            if mesh_vars.CGNS.regenerate_BCs:
                hopout.separator()
                hopout.info('REGENERATE MESH...')
                mesh = BCCGNS()
                hopout.info('REGENERATE MESH DONE!')
            else:
                mesh = mesh_vars.mesh
        case _:  # Default
            hopout.warning('Unknown mesh mode {}, exiting...'.format(mesh_vars.mode))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)

    mesh_vars.mesh = mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
from typing import Tuple
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def facePointMatrixFill(matrix: np.ndarray, start: int, end: int, count: int, orient: bool) -> Tuple[np.ndarray, int]:
    """ Fill the 2D matrix representing the inner points of each faces
    """
    if end <= start:
        return matrix, count

    # Fill the corner nodes first
    matrix[start, start] = count
    count += 1
    if orient:
        matrix[start, end  ] = count
        count += 1
        matrix[end  , end  ] = count
        count += 1
        matrix[end  , start] = count
        count += 1
    else:
        matrix[end  , start] = count
        count += 1
        matrix[end  , end  ] = count
        count += 1
        matrix[start, end  ] = count
        count += 1

    # Now, fill the edges
    if orient:
        for i in range(start+1, end):
            matrix[start, i    ] = count
            count += 1
        for i in range(start+1, end):
            matrix[i    , end  ] = count
            count += 1
        for i in range(end-1, start, -1):
            matrix[end  , i    ] = count
            count += 1
        for i in range(end-1, start, -1):
            matrix[i    , start] = count
            count += 1
    else:
        for i in range(start+1, end):
            matrix[i    , start] = count
            count += 1
        for i in range(start+1, end):
            matrix[end  , i    ] = count
            count += 1
        for i in range(end-1, start, -1):
            matrix[i    , end  ] = count
            count += 1
        for i in range(end-1, start, -1):
            matrix[start, i    ] = count
            count += 1

    return matrix, count


def facePointMatrix(order: int, pos: int, orient: bool = True) -> np.ndarray:
    """ Return the 2D index of the inner points of each faces
    """
    # Create a matrix of the required size
    matrix = np.zeros((order, order), dtype=int)

    count  = 0
    # Fill the matrix recursively
    for i in range(np.floor(order/2).astype(int)):
        matrix, count = facePointMatrixFill(matrix, i, order-i-1, count, orient)

    # Fill the middle point if uneven order
    if order % 2 != 0:
        matrix[np.floor(order/2).astype(int), np.floor(order/2).astype(int)] = count
    return np.argwhere(matrix == pos)[0]


def edgePointMESHIO(start: int, end: int, edge: int, node: int) -> np.ndarray:
    """ Traverse over all 12 edges of the hexahedron
    """
    match edge:
        case 0:
            return np.array([node          , start , start  ], dtype=int)
        case 1:
            return np.array([start         , node  , start  ], dtype=int)
        case 2:
            return np.array([start         , start , node   ], dtype=int)
        case 3:
            return np.array([end           , node  , start  ], dtype=int)
        case 4:
            return np.array([end           , start , node   ], dtype=int)
        case 5:
            return np.array([end+start-node, end   , start  ], dtype=int)
        case 6:
            return np.array([end           , end   , node   ], dtype=int)
        case 7:
            return np.array([start         , end   , node   ], dtype=int)
        case 8:
            return np.array([node          , start , end    ], dtype=int)
        case 9:
            return np.array([start         , node  , end    ], dtype=int)
        case 10:
            return np.array([end           , node  , end    ], dtype=int)
        case 11:
            return np.array([end+start-node, end   , end    ], dtype=int)
        case _:
            sys.exit(1)


def facePointMESHIO(start: int, end: int, face: int, pos: int) -> np.ndarray:
    """ Translate the 1D position of each of the 6 hexahedron faces to each 2D index
    """
    match face:
        case 0:
            index = facePointMatrix(end-start-1, pos, orient=True)
            return np.array([start+index[0]+1 , start+index[1]+1 , start           ], dtype=int)
        case 1:
            index = facePointMatrix(end-start-1, pos, orient=False)
            return np.array([start+index[0]+1 , start            , start+index[1]+1], dtype=int)
        case 2:
            index = facePointMatrix(end-start-1, pos, orient=True)
            return np.array([start            , start+index[0]+1 , start+index[1]+1], dtype=int)
        case 3:
            index = facePointMatrix(end-start-1, pos, orient=False)
            return np.array([end              , start+index[0]+1 , start+index[1]+1], dtype=int)
        case 4:
            index = facePointMatrix(end-start-1, pos, orient=False)
            return np.array([end  -index[0]-1 , end              , start+index[1]+1], dtype=int)
        case 5:
            index = facePointMatrix(end-start-1, pos, orient=False)
            return np.array([start+index[0]+1 , start+index[1]+1 , end             ], dtype=int)
        case _:
            sys.exit(1)


def genHEXMAPMESHIO(order: int) -> None:
    """ MESHIO -> IJK ordering for high-order hexahedrons
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    map = np.zeros((order, order, order), dtype=int)

    if order == 1:
        map[0, 0, 0] = 0
        mesh_vars.HEXMAP = map
        return None

    if order == 2:
        # Python indexing, 1 -> 0
        map -= 1
        # Reshape into 1D array, tensor-product style
        tensor = []
        for k in range(order):
            for j in range(order):
                for i in range(order):
                    tensor.append(int(map[i, j, k]))

        mesh_vars.HEXMAP = tensor
        return None

    count = 0

    # Fill the cube recursively from the outside to the inside
    for iOrder in range(np.floor(order/2).astype(int)):
        # Principal vertices
        map[iOrder         , iOrder         , iOrder        ] = count+1
        map[order-iOrder-1 , iOrder         , iOrder        ] = count+2
        map[order-iOrder-1 , order-iOrder-1 , iOrder        ] = count+3
        map[iOrder         , order-iOrder-1 , iOrder        ] = count+4
        map[iOrder         , iOrder         , order-iOrder-1] = count+5
        map[order-iOrder-1 , iOrder         , order-iOrder-1] = count+6
        map[order-iOrder-1 , order-iOrder-1 , order-iOrder-1] = count+7
        map[iOrder         , order-iOrder-1 , order-iOrder-1] = count+8
        count += 8

        # Loop over all edges
        for iEdge in range(12):
            for iNode in range(iOrder+1, order-iOrder-1):
                # Assemble mapping to tuple
                count += 1
                edge  = edgePointMESHIO(iOrder, order-iOrder-1, iEdge, iNode)
                index = (int(edge[0]), int(edge[1]), int(edge[2]))
                map[index] = count

        # Internal points of upstanding faces
        for iFace in range(6):
            for pos in range((order-2*iOrder-2)**2):
                # Assemble mapping to tuple, top  quadrangle -> z = order
                count += 1
                index = facePointMESHIO(iOrder, order-iOrder-1, iFace, pos)
                index = (int(index[0]), int(index[1]), int(index[2]))
                map[index] = count

    # Fill the middle point if uneven order
    if order % 2 != 0:
        index = (np.floor(order/2).astype(int), np.floor(order/2).astype(int), np.floor(order/2).astype(int))
        map[index] = count+1

    # Python indexing, 1 -> 0
    map -= 1

    # Reshape into 1D array, tensor-product style
    tensor = []
    for k in range(order):
        for j in range(order):
            for i in range(order):
                tensor.append(int(map[i, j, k]))

    mesh_vars.HEXMAP = tensor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def edgePointVTK(order: int, edge: int, node: int) -> np.ndarray:
    match edge:
        case 0:  # z- / base
            return np.array([node      , 0         ], dtype=int)
        case 1:  # y+ / base
            return np.array([order     , node      ], dtype=int)
        case 2:  # z+ / base
            return np.array([order-node, order     ], dtype=int)
        case 3:  # y- / base
            return np.array([0         , order-node], dtype=int)
        case _:
            sys.exit(1)


def genHEXMAPVTK(order: int) -> None:
    """ VTK -> IJK ordering for high-order hexahedrons
        > Losely based on [Gmsh] "generatePointsHexCGNS"
        > [Jens Ulrich Kreber] "paraview-scripts/node_ordering.py"
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    map = np.zeros((order, order, order), dtype=int)

    if order == 1:
        map[0, 0, 0] = 0
        mesh_vars.HEXMAP = map
        return None

    # Principal vertices
    map[0      , 0      , 0      ] = 1
    map[order-1, 0      , 0      ] = 2
    map[order-1, order-1, 0      ] = 3
    map[0      , order-1, 0      ] = 4
    map[0      , 0      , order-1] = 5
    map[order-1, 0      , order-1] = 6
    map[order-1, order-1, order-1] = 7
    map[0      , order-1, order-1] = 8

    if order == 2:
        # Python indexing, 1 -> 0
        map -= 1
        # Reshape into 1D array, tensor-product style
        tensor = []
        for k in range(order):
            for j in range(order):
                for i in range(order):
                    tensor.append(int(map[i, j, k]))

        mesh_vars.HEXMAP = tensor
        return None

    # Internal points of base quadrangle edges (x-)
    count = 8
    for iFace in range(4):
        for iNode in range(1, order-1):
            # Assemble mapping to tuple, base quadrangle -> z = 0
            count += 1
            edge  = edgePointVTK(order-1, iFace, iNode)
            index = (int(edge[0]), int(edge[1]), 0)
            map[index] = count

    # Internal points of top quadrangle edges
    for iFace in range(4):
        for iNode in range(1, order-1):
            # Assemble mapping to tuple, top  quadrangle -> z = order
            count += 1
            edge  = edgePointVTK(order-1, iFace, iNode)
            index = (int(edge[0]), int(edge[1]), order-1)
            map[index] = count

    # Internal points of mounting edges
    for iFace in range(4):
        edge  = edgePointVTK(order-1, (iFace+3) % 4, order-1)
        for iNode in range(1, order-1):
            # Assemble mapping to tuple, mounting edges -> z ascending
            count += 1
            index = (int(edge[0]), int(edge[1]), iNode)
            map[index] = count
    # > VTK9 swapped 3/4
    # (0      ,0)
    # for iNode in range(1, order-1):
    #     # Assemble mapping to tuple, mounting edges -> z ascending
    #     count += 1
    #     index = (0           , 0           , iNode)
    #     map[index] = count
    # # (order-1,0)
    # for iNode in range(1, order-1):
    #     # Assemble mapping to tuple, mounting edges -> z ascending
    #     count += 1
    #     index = (order-1     , 0           , iNode)
    #     map[index] = count
    # # (order-1,order-1)
    # for iNode in range(1, order-1):
    #     # Assemble mapping to tuple, mounting edges -> z ascending
    #     count += 1
    #     index = (order-1     , order-1     , iNode)
    #     map[index] = count
    # # (order-1,order-1)
    # for iNode in range(1, order-1):
    #     # Assemble mapping to tuple, mounting edges -> z ascending
    #     count += 1
    #     index = (order-1     , order-1     , iNode)
    #     map[index] = count

    # WARNING:THIS IS HOW IT SHOULD BE
    # Internal points of upstanding faces
    # > x- face
    k = 0
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            count += 1
            index = (k    , i+1  , j+1  )
            map[index] = count
    # > x+ face
    k = order-1
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            count += 1
            index = (k    , i+1  , j+1  )
            map[index] = count
    # > y- face
    k = 0
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            count += 1
            index = (i+1  , k    , j+1  )
            map[index] = count
    # > y+ face
    k = order-1
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            count += 1
            index = (i+1  , k    , j+1  )
            map[index] = count
    # Internal points of base quadrangle (z-)
    k = 0
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            count += 1
            index = (i+1  , j+1  , k    )
            map[index] = count

    # Internal points of top  quadrangle (z+)
    k = order-1
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            count += 1
            index = (i+1  , j+1  , k    )
            map[index] = count

    # # FIXME: THIS IS HOW MESHIO GIVES IT
    # # Internal points of base quadrangle (z-)
    # k = 0
    # # Fill the map
    # for j in range(order-2):
    #     for i in range(order-2):
    #         count += 1
    #         index = (i+1  , j+1  , k    )
    #         map[index] = count
    # # > y- face
    # k = 0
    # # Fill the map
    # for j in range(order-2):
    #     for i in range(order-2):
    #         count += 1
    #         index = (i+1  , k    , j+1  )
    #         map[index] = count
    # # > x+ face
    # k = order-1
    # # Fill the map
    # for j in range(order-2):
    #     for i in range(order-2):
    #         count += 1
    #         index = (k    , i+1  , j+1  )
    #         map[index] = count
    # # > y+ face
    # k = order-1
    # # Fill the map
    # for j in range(order-2):
    #     for i in range(order-2):
    #         count += 1
    #         index = (i+1  , k    , j+1  )
    #         map[index] = count
    # # > x- face
    # k = 0
    # # Fill the map
    # for j in range(order-2):
    #     for i in range(order-2):
    #         count += 1
    #         index = (k    , i+1  , j+1  )
    #         map[index] = count
    # # Internal points of top  quadrangle (z+)
    # k = order-1
    # # Fill the map
    # for j in range(order-2):
    #     for i in range(order-2):
    #         count += 1
    #         index = (i+1  , j+1  , k    )
    #         map[index] = count

    # Internal volume points as a tensor product
    for k in range(1, order-1):
        for j in range(1, order-1):
            for i in range(1, order-1):
                count += 1
                index = (i  , j  , k  )
                map[index] = count

    # Python indexing, 1 -> 0
    map -= 1

    # Reshape into 1D array, tensor-product style
    tensor = []
    for k in range(order):
        for j in range(order):
            for i in range(order):
                tensor.append(int(map[i, j, k]))
    mesh_vars.HEXMAP = tensor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import argparse
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
STD_LENGTH = 79  # Standard length for output to console
PAR_LENGTH = 32
DEF_LENGTH = 20
# ==================================================================================================================================


def info(string: str, newline: bool = False) -> None:
    if newline:
        print('\n!', string)
    else:
        print('!', string)


def separator(length: int = STD_LENGTH) -> str:
    return '!' + '-'*(length-1)


class CommandLine:
    """ Parse command line arguments, both explicit [*.ini] and flags [--]
    """
    def __init__(self, argv, name: str, version: str, commit: str) -> None:
        # Local imports ----------------------------------------
        import pyhope.config.config as config
        from pyhope.output.output import Colors
        # ------------------------------------------------------

        # Read the command line arguments and store everything
        self.argv    = argv
        self.name    = name
        self.version = version
        self.commit  = commit
        # self.help    = ''

        # Print the header
        self.help = (Colors.BANNERA + '!' + '='*(STD_LENGTH-1))
        self.helpjoin('! {} version {} [commit {}]'.format(name, version, commit))
        self.helpjoin(Colors.BANNERA + '!' + '='*(STD_LENGTH-1) + Colors.END)

        # Assemble the help output
        for key in config.prms:
            # Check if we encountered a section
            if config.prms[key]['type'] == 'section':
                self.helpjoin(separator())
                self.helpjoin('! {}'.format(key))
                self.helpjoin(separator())
                continue

            if config.prms[key]['default']:
                default = config.prms[key]['default']
            else:
                default = ''

            if config.prms[key]['help']:
                help    = config.prms[key]['help']
            else:
                help    = ''

            self.helpjoin(f'{key:<{PAR_LENGTH}} = {default:>{DEF_LENGTH}} ! {help}')

        return None

    def __enter__(self):
        # Setup an argument parser and add know arguments
        parser = argparse.ArgumentParser(prog=self.name,
                                         formatter_class=argparse.RawDescriptionHelpFormatter,
                                         epilog=self.help)
        parser.add_argument('-V', '--version',
                            action='store_true',
                            help='display the version number and exit')
        parser.add_argument('parameter',
                            nargs='?',
                            metavar='<parameter.ini>',
                            help='HOPR parameter file')
        # Parse known arguments and return other flags for further processing
        args, argv = parser.parse_known_args()
        return args, argv

    def __exit__(self, *args: object) -> None:
        return None

    def helpjoin(self, end) -> None:
        self.help = os.linesep.join([self.help, end])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import string
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def GenerateSides() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.mesh.mesh_common import face_to_cgns, faces
    # ------------------------------------------------------

    mesh   = mesh_vars.mesh
    nElems = 0
    nSides = 0
    sCount = 0
    mesh_vars.elems = []
    mesh_vars.sides = []
    elems   = mesh_vars.elems
    sides   = mesh_vars.sides

    hopout.sep()
    hopout.routine('Eliminating duplicate points')

    # Eliminate duplicate points
    mesh_vars.mesh.points, inverseIndices = np.unique(mesh_vars.mesh.points, axis=0, return_inverse=True)

    # Update the mesh
    for cell in mesh_vars.mesh.cells:
        # Map the old indices to the new ones
        # cell.data = np.vectorize(lambda idx: inverseIndices[idx])(cell.data)
        # Efficiently map all indices in one operation
        cell.data = inverseIndices[cell.data]

    # Loop over all element types
    for iType, elemType in enumerate(mesh.cells_dict.keys()):
        # Only consider three-dimensional types
        if not any(s in elemType for s in mesh_vars.ELEM.type.keys()):
            continue

        # Get the elements
        ioelems  = mesh.get_cells_type(elemType)
        baseElem = elemType.rstrip(string.digits)
        nIOElems = ioelems.shape[0]
        nIOSides   = mesh_vars.ELEM.type[baseElem]

        # Create non-unique sides
        mesh_vars.elems.extend([dict() for _ in range(nIOElems         )])
        mesh_vars.sides.extend([dict() for _ in range(nIOElems*nIOSides)])

        # Create dictionaries
        for iElem in range(nElems, nElems+nIOElems):
            elems[iElem]['Type'  ] = mesh_vars.ELEMMAP(elemType)
            elems[iElem]['ElemID'] = iElem
            elems[iElem]['Sides' ] = []
            elems[iElem]['Nodes' ] = ioelems[iElem]

            # Create the sides
            for iSide in range(nSides, nSides+nIOSides):
                sides[iSide]['Type'  ] = 4  # FIXME: THIS NEEDS TREATMENT FOR NON-HEXAS

            # Assign nodes to sides, CGNS format
            for index, face in enumerate(faces()):
                corners = [ioelems[iElem][s] for s in face_to_cgns(face)]
                sides[sCount].update({'ElemID' : iElem})
                sides[sCount].update({'SideID' : sCount})
                sides[sCount].update({'LocSide': index+1})
                sides[sCount].update({'Corners': np.array(corners)})
                sCount += 1

            # Add to nSides
            nSides += nIOSides

        # Add to nElems
        nElems += nIOElems

    # Append sides to elem
    for iSide, side in enumerate(sides):
        elemID = side['ElemID']
        sideID = side['SideID']
        elems[elemID]['Sides'].append(sideID)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
from typing import Tuple
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
import h5py
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


class ELEM:
    INFOSIZE  = 6
    TYPE      = 0
    ZONE      = 1
    FIRSTSIDE = 2
    LASTSIDE  = 3
    FIRSTNODE = 4
    LASTNODE  = 5

    TYPES     = [104, 204, 105, 115, 205, 106, 116, 206, 108, 118, 208]


class SIDE:
    INFOSIZE  = 5
    TYPE      = 0
    ID        = 1
    NBELEMID  = 2
    NBLOCSIDE_FLIP = 3
    BCID      = 4


def LINMAP(elemType: int, order: int = 1) -> np.ndarray:
    """ CGNS -> IJK ordering for element corner nodes
    """
    # Local imports ----------------------------------------
    # from pyhope.io.io_cgns import genHEXMAPCGNS
    # from pyhope.io.io_vtk import genHEXMAPVTK
    from pyhope.io.io_meshio import genHEXMAPMESHIO
    # ------------------------------------------------------
    match elemType:
        # Straight-sided elements, hard-coded
        case 104:  # Tetraeder
            return np.array([0, 1, 2, 3])
        case 105:  # Pyramid
            return np.array([0, 1, 3, 2, 4])
        case 106:  # Prism
            return np.array([0, 1, 2, 3, 4, 5])
        case 108:  # Hexaeder
            return np.array([0, 1, 3, 2, 4, 5, 7, 6])
        # Curved elements, use mapping
        case 208:  # Hexaeder
            # # CGNS
            # try:
            #     from pyhope.mesh.mesh_vars import HEXMAP
            # except ImportError:
            #     genHEXMAP(order+1)
            #     from pyhope.mesh.mesh_vars import HEXMAP

            # # VTK
            # try:
            #     from pyhope.mesh.mesh_vars import HEXMAP
            # except ImportError:
            #     genHEXMAPVTK(order+1)
            #     from pyhope.mesh.mesh_vars import HEXMAP

            # MESHIO
            try:
                from pyhope.mesh.mesh_vars import HEXMAP
            except ImportError:
                genHEXMAPMESHIO(order+1)
                from pyhope.mesh.mesh_vars import HEXMAP
            return HEXMAP
        case _:  # Default
            print('Error in LINMAP, unknown elemType')
            sys.exit(1)


def ELEMTYPE(elemType: int) -> str:
    """ Name of a given element type
    """
    match elemType:
        case 104:
            return ' Straight-edge Tetrahedra '
        case 204:
            return '        Curved Tetrahedra '
        case 105:
            return '  Planar-faced Pyramids   '
        case 115:
            return ' Straight-edge Pyramids   '
        case 205:
            return '        Curved Pyramids   '
        case 106:
            return '  Planar-faced Prisms     '
        case 116:
            return ' Straight-edge Prisms     '
        case 206:
            return '        Curved Prisms     '
        case 108:
            return '  Planar-faced Hexahedra  '
        case 118:
            return ' Straight-edge Hexahedra  '
        case 208:
            return '        Curved Hexahedra  '
        case _:  # Default
            print('Error in ELEMTYPE, unknown elemType')
            sys.exit(1)


def DefineIO() -> None:
    # Local imports ----------------------------------------
    from pyhope.io.io_vars import MeshFormat
    from pyhope.readintools.readintools import CreateIntFromString, CreateIntOption, CreateLogical, CreateSection, CreateStr
    # ------------------------------------------------------

    CreateSection('Output')
    CreateStr('ProjectName', help='Name of output files')
    CreateIntFromString('OutputFormat', default=0    , help='Mesh output format')
    CreateIntOption(    'OutputFormat', number=MeshFormat.FORMAT_HDF5, name='HDF5')
    CreateIntOption(    'OutputFormat', number=MeshFormat.FORMAT_VTK , name='VTK')
    CreateLogical(      'DebugVisu'   , default=False, help='Launch the GMSH GUI to visualize the mesh')


def InitIO() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import GetIntFromStr, GetLogical, GetStr
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('INIT OUTPUT...')

    io_vars.projectname  = GetStr('ProjectName')
    io_vars.outputformat = GetIntFromStr('OutputFormat')

    io_vars.debugvisu    = GetLogical('DebugVisu')

    hopout.info('INIT OUTPUT DONE!')


def IO() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common_vars import Common
    from pyhope.io.io_vars import MeshFormat
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('OUTPUT MESH...')

    match io_vars.outputformat:
        case MeshFormat.FORMAT_HDF5:
            mesh  = mesh_vars.mesh
            elems = mesh_vars.elems
            sides = mesh_vars.sides

            nElems = len(elems)
            nSides = len(sides)
            nNodes = np.sum([s['Nodes'].size for s in elems])  # number of non-unique nodes

            bcs   = mesh_vars.bcs
            nBCs  = len(bcs)

            pname = io_vars.projectname
            fname = '{}_mesh.h5'.format(pname)

            # elemInfo, sideInfo, nodeInfo, nodeCoords, elemCounter = getMeshInfo()
            elemInfo, sideInfo, _, nodeCoords, elemCounter = getMeshInfo()

            # Print the final output
            hopout.sep()
            for elemType in ELEM.TYPES:
                if elemCounter[elemType] > 0:
                    hopout.info( ELEMTYPE(elemType) + ': {:12d}'.format(elemCounter[elemType]))
            hopout.sep()
            hopout.routine('Writing HDF5 mesh to "{}"'.format(fname))
            hopout.sep()

            with h5py.File(fname, mode='w') as f:
                # Store same basic information
                common = Common()
                f.attrs['HoprVersion'   ] = common.version
                f.attrs['HoprVersionInt'] = common.__version__.micro + common.__version__.minor*100 + common.__version__.major*10000

                # Store mesh information
                f.attrs['Ngeo'          ] = mesh_vars.nGeo
                f.attrs['nElems'        ] = nElems
                f.attrs['nSides'        ] = nSides
                f.attrs['nNodes'        ] = nNodes

                f.create_dataset('ElemInfo'  , data=elemInfo)
                f.create_dataset('SideInfo'  , data=sideInfo)
                # f.create_dataset('NodeInfo'  , data=nodeInfo)
                f.create_dataset('NodeCoords', data=nodeCoords)

                # Store boundary information
                f.attrs['nBCs'          ] = nBCs
                bcNames = [f'{s["Name"]:<255}' for s in bcs]
                bcTypes = np.zeros((nBCs, 4), dtype=np.int32)
                for iBC, bc in enumerate(bcs):
                    bcTypes[iBC, :] = bc['Type']

                f.create_dataset('BCNames'   , data=np.bytes_(bcNames))
                f.create_dataset('BCType'    , data=bcTypes)

        case MeshFormat.FORMAT_VTK:
            mesh  = mesh_vars.mesh
            pname = io_vars.projectname
            fname = '{}_mesh.vtk'.format(pname)

            hopout.routine('Writing VTK mesh to "{}"'.format(fname))

            mesh.write(fname, file_format='vtk42')

        case _:  # Default
            hopout.warning('Unknown output format {}, exiting...'.format(io_vars.outputformat))
            sys.exit(1)

    hopout.info('OUTPUT MESH DONE!')


def getMeshInfo() -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, dict]:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------

    mesh  = mesh_vars.mesh
    elems = mesh_vars.elems
    sides = mesh_vars.sides
    nodes = mesh.points

    nElems = len(elems)
    nSides = len(sides)
    nNodes = np.sum([s['Nodes'].size for s in elems])  # number of non-unique nodes

    # Create the ElemCounter
    elemCounter = dict()
    for iType, elemType in enumerate(ELEM.TYPES):
        elemCounter[elemType] = 0

    # Fill the ElemInfo
    elemInfo  = np.zeros((nElems, ELEM.INFOSIZE), dtype=np.int32)
    sideCount = 0  # elem['Sides'] might work as well
    nodeCount = 0  # elem['Nodes'] contains the unique nodes

    for iElem, elem in enumerate(elems):
        elemInfo[iElem, ELEM.TYPE     ] = elem['Type']
        elemInfo[iElem, ELEM.ZONE     ] = 1  # FIXME

        elemInfo[iElem, ELEM.FIRSTSIDE] = sideCount
        elemInfo[iElem, ELEM.LASTSIDE ] = sideCount + len(elem['Sides'])
        sideCount += len(elem['Sides'])

        elemInfo[iElem, ELEM.FIRSTNODE] = nodeCount
        elemInfo[iElem, ELEM.LASTNODE ] = nodeCount + len(elem['Nodes'])
        nodeCount += len(elem['Nodes'])

        elemCounter[elem['Type']] += 1

    # Fill the SideInfo
    sideInfo  = np.zeros((nSides, SIDE.INFOSIZE), dtype=np.int32)

    for iSide, side in enumerate(sides):
        sideInfo[iSide, SIDE.TYPE     ] = side['Type'  ]
        sideInfo[iSide, SIDE.ID       ] = side['GlobalSideID']
        # Connected sides
        if 'Connection' in side:
            nbSideID = side['Connection']
            nbElemID = sides[nbSideID]['ElemID'] + 1  # Python -> HOPR index
            sideInfo[iSide, SIDE.NBELEMID      ] = nbElemID
            if side['Flip'] == 0:  # Master side
                sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = sides[nbSideID]['LocSide']*10
            else:
                sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = sides[nbSideID]['LocSide']*10 + side['Flip']

            # Periodic sides still have a BCID
            if 'BCID' in side:
                sideInfo[iSide, SIDE.BCID      ] = side['BCID'] + 1
            else:
                sideInfo[iSide, SIDE.BCID      ] = 0
        else:
            sideInfo[iSide, SIDE.NBELEMID      ] = 0
            sideInfo[iSide, SIDE.NBLOCSIDE_FLIP] = 0
            sideInfo[iSide, SIDE.BCID          ] = side['BCID'] + 1

    # Fill the NodeInfo
    nodeInfo = np.zeros((ELEM.INFOSIZE, nNodes), dtype=np.int32)

    # Fill the NodeCoords
    nodeCoords = np.zeros((nNodes, 3), dtype=np.float64)
    nodeCount  = 0

    for iElem, elem in enumerate(elems):
        # Mesh coordinates are stored in meshIO sorting
        linMap    = LINMAP(elem['Type'], order=mesh_vars.nGeo)
        # meshio accesses them in their own ordering
        # > need to reverse the mapping
        mapLin    = {k: v for v, k in enumerate(linMap)}
        elemNodes = elem['Nodes']

        # Access the actual nodeCoords and reorder them
        for iNode, nodeID in enumerate(elemNodes):
            nodeCoords[nodeCount + mapLin[iNode], :] = nodes[nodeID]

        nodeCount += len(elemNodes)

    return elemInfo, sideInfo, nodeInfo, nodeCoords, elemCounter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from multiprocessing import Pool, Queue
import plotext as plt
from alive_progress import alive_bar
from typing import Tuple
import threading
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def legendre_gauss_lobatto_nodes(order: int) -> Tuple[np.ndarray, np.ndarray]:
    """ Return Legendre-Gauss-Lobatto nodes and weights for a given order in 1D
    """
    order -= 1
    # Special cases for small N
    if order == 1:
        return np.array([-1, 1]), np.array([1, 1])

    # Compute the initial guess for the LGL nodes (roots of P'_N)
    nodes = np.cos(np.pi * np.arange(order+1) / order)

    # Initialize the Legendre polynomial and its derivative
    p = np.zeros((order+1, order+1))

    # Iteratively solve for the LGL nodes using Newton's method
    xOld = 2 * np.ones_like(nodes)
    tol = 1e-14
    while np.max(np.abs(nodes - xOld)) > tol:
        xOld = nodes.copy()
        p[:, 0] = 1
        p[:, 1] = nodes
        for k in range(2, order+1):
            p[:, k] = ((2*k-1) * nodes * p[:, k-1] - (k-1) * p[:, k-2]) / k
        nodes -= (nodes * p[:, order] - p[:, order-1]) / (order * (p[:, order]))

    # The LGL nodes
    nodes = np.sort(nodes)

    # Compute the LGL weights
    weights = 2 / (order * (order + 1) * (p[:, order]**2))

    return nodes, weights


def legendre_gauss_nodes(order: int) -> Tuple[np.ndarray, np.ndarray]:
    """ Return Legendre-Gauss nodes and weights for a given order in 1D
    """
    nodes, weights = np.polynomial.legendre.leggauss(order)
    return nodes, weights


def barycentric_weights(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the barycentric weights for a given node set
        > Algorithm 30, Kopriva
    """
    # Create a difference matrix (x_i - x_j) for all i, j
    diff_matrix = xGP[:, np.newaxis] - xGP[np.newaxis, :]

    # Set the diagonal to 1 to avoid division by zero (diagonal elements will not be used)
    np.fill_diagonal(diff_matrix, 1.0)

    # Compute the product of all differences for each row (excluding the diagonal)
    wBary = np.prod(diff_matrix, axis=1)

    # Take the reciprocal to get the final barycentric weights
    wBary = 1.0 / wBary

    return wBary


def polynomial_derivative_matrix(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the polynomial derivative matrix for a given node set
        > Algorithm 37, Kopriva
    """
    wBary = barycentric_weights(order, xGP)
    D     = np.zeros((order, order), dtype=float)

    for iLagrange in range(order):
        for iGP in range(order):
            if iLagrange != iGP:
                D[iGP, iLagrange] = wBary[iLagrange]/(wBary[iGP]*(xGP[iGP]-xGP[iLagrange]))
                D[iGP, iGP      ] = D[iGP, iGP] - D[iGP, iLagrange]

    return D


def lagrange_interpolation_polys(x: float, order: int, xGP: np.ndarray, wBary: np.ndarray) -> np.ndarray:
    """ Computes all Lagrange functions evaluated at position x in [-1;1]
        > Algorithm 34, Kopriva
    """
    # Equal points need special treatment
    L   = np.zeros(order)
    for iGP in range(order):
        if abs(x - xGP[iGP]) < 1.E-14:
            L[iGP] = 1
            return L

    tmp = 0.
    for iGP in range(order):
        L[iGP] = wBary[iGP] / (x-xGP[iGP])
        tmp   += L[iGP]

    # Normalize
    L = L/tmp
    return L


def calc_vandermonde(n_In: int, n_Out: int, wBary_In: np.ndarray, xi_In: np.ndarray, xi_Out: np.ndarray) -> np.ndarray:
    """ Build a 1D Vandermonde matrix using the Lagrange basis functions of degree N_In,
        evaluated at the interpolation points xi_Out
    """
    Vdm = np.zeros((n_Out, n_In))
    for iXI in range(n_Out):
        Vdm[iXI, :] = lagrange_interpolation_polys(xi_Out[iXI], n_In, xi_In, wBary_In)
    return Vdm


# def change_basis_3D(dim1: int, n_In: int, n_Out: int, Vdm: np.ndarray, x3D_In: np.ndarray) -> np.ndarray:
def change_basis_3D(Vdm: np.ndarray, x3D_In: np.ndarray) -> np.ndarray:
    """ Interpolate a 3D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
        to another 3D tensor product node positions (number of nodes N_out+1)
        defined by (N_out+1) interpolation point  positions xi_Out(0:N_Out)
        xi is defined in the 1DrefElem xi=[-1,1]
    """
    # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x3D_In)
    x3D_Buf1 = np.tensordot(Vdm, x3D_In, axes=(1, 1))
    x3D_Buf1 = np.moveaxis(x3D_Buf1, 0, 1)  # Correct the shape to (dim1, n_Out, n_In, n_In)

    # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of x3D_Buf1)
    x3D_Buf2 = np.tensordot(Vdm, x3D_Buf1, axes=(1, 2))
    x3D_Buf2 = np.moveaxis(x3D_Buf2, 0, 2)  # Correct the shape to  (dim1, n_Out, n_Out, n_In)

    # Third contraction along the kN_In axis (axis 1 of Vdm, axis 3 of x3D_Buf2)
    x3D_Out  = np.tensordot(Vdm, x3D_Buf2, axes=(1, 3))
    x3D_Out  = np.moveaxis(x3D_Out , 0, 1)  # Correct the shape to (dim1, n_Out, n_Out, n_Out)

    return x3D_Out


def evaluate_jacobian(xGeo_In: np.ndarray, VdmGLtoAP: np.ndarray, D_EqToGL: np.ndarray) -> np.ndarray:
    # Perform tensor contraction for the first derivative (Xi direction)
    dXdXiGL   = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 1))
    dXdXiGL   = np.moveaxis(dXdXiGL  , 0, -3)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)

    # Perform tensor contraction for the second derivative (Eta direction)
    dXdEtaGL  = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 2))
    dXdEtaGL  = np.moveaxis(dXdEtaGL , 0, -3)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)

    # Perform tensor contraction for the third derivative (Zeta direction)
    dXdZetaGL = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 3))
    dXdZetaGL = np.moveaxis(dXdZetaGL, 0, -3)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)

    # Change basis for each direction
    dXdXiAP   = change_basis_3D(VdmGLtoAP, dXdXiGL  )
    dXdEtaAP  = change_basis_3D(VdmGLtoAP, dXdEtaGL )
    dXdZetaAP = change_basis_3D(VdmGLtoAP, dXdZetaGL)

    # Precompute cross products between dXdEtaAP and dXdZetaAP for all points
    cross_eta_zeta = np.cross(dXdEtaAP, dXdZetaAP, axis=0)  # Shape: (3, nGeoRef, nGeoRef, nGeoRef)

    # Fill output Jacobian array
    jacOut = np.einsum('ijkl,ijkl->jkl', dXdXiAP, cross_eta_zeta)

    return jacOut


def plot_histogram(data: np.ndarray) -> None:
    """ Plot a histogram of all Jacobians
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.output.output import STD_LENGTH
    # ------------------------------------------------------

    ticks = ['│<0.0      │',
             '│ 0.0-0.1  │',
             '│ 0.1-0.2  │',
             '│ 0.2-0.3  │',
             '│ 0.3-0.4  │',
             '│ 0.4-0.5  │',
             '│ 0.5-0.6  │',
             '│ 0.6-0.7  │',
             '│ 0.7-0.8  │',
             '│ 0.8-0.9  │',
             # '│ 0.9-0.99 │',
             '│>0.9-1.0  │']

    # Define the bins for categorizing jacobians
    # bins     = [ -np.inf, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.99, np.inf]
    bins     = [ -np.inf, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9,       np.inf]

    # Use np.histogram to count jacobians in the defined bins
    count, _ = np.histogram(data, bins=bins)

    # Setup plot
    hopout.sep()
    hopout.info('Scaled Jacobians')
    hopout.separator(18)
    plt.simple_bar(ticks, count, width=STD_LENGTH)
    plt.show()
    hopout.separator(18)


def process_chunk(chunk):
    """Process a chunk of elements by evaluating the Jacobian for each
    """
    chunk_results = []
    for elem in chunk:
        # nodeCoords, nGeoRef, VdmGLtoAP, D_EqToGL = elem
        nodeCoords, _, VdmGLtoAP, D_EqToGL = elem
        jac    = evaluate_jacobian(nodeCoords, VdmGLtoAP, D_EqToGL)
        maxJac = np.max(np.abs(jac))
        minJac = np.min(jac)
        chunk_results.append(minJac / maxJac)
    return chunk_results


def distribute_work(elems, chunk_size):
    """Distribute elements into chunks of a given size
    """
    return [elems[i:i + chunk_size] for i in range(0, len(elems), chunk_size)]


def run_in_parallel(elems, chunk_size=10):
    """Run the element processing in parallel using a specified number of processes
    """
    # Local imports ----------------------------------------
    from pyhope.common.common_vars import np_mtp
    # ------------------------------------------------------

    chunks = distribute_work(elems, chunk_size)
    total_elements = len(elems)
    progress_queue = Queue()

    # Use a separate thread for the progress bar
    progress_thread = threading.Thread(target=update_progress, args=(progress_queue, total_elements))
    progress_thread.start()

    # Use multiprocessing Pool for parallel processing
    with Pool(processes=np_mtp) as pool:
        # Map work across processes in chunks
        results = []
        for chunk_result in pool.imap_unordered(process_chunk, chunks):
            results.extend(chunk_result)
            # Update progress for each processed element in the chunk
            for _ in chunk_result:
                progress_queue.put(1)

    # Wait for the progress bar thread to finish
    progress_thread.join()

    return results


def update_progress(progress_queue, total_elements):
    """ Function to update the progress bar from the queue
    """
    with alive_bar(total_elements, title='│             Processing Elements', length=33) as bar:
        for _ in range(total_elements):
            # Block until we receive a progress update from the queue
            progress_queue.get()
            bar()


def CheckJacobians() -> None:
    # Local imports ----------------------------------------
    from pyhope.common.common_vars import np_mtp
    from pyhope.io.io import LINMAP
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import GetLogical
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('CHECK JACOBIANS...')
    hopout.sep()

    checkElemJacobians = GetLogical('CheckElemJacobians')
    if not checkElemJacobians:
        return None

    nGeo = mesh_vars.nGeo + 1

    # Compute the equidistant point set used by meshIO
    xEq = np.zeros(nGeo)
    for i in range(nGeo):
        xEq[i] = 2.*float(i)/float(nGeo-1) - 1.
    wBaryEq   = barycentric_weights(nGeo, xEq)

    xGL, _    = legendre_gauss_lobatto_nodes(nGeo)
    DGL       = polynomial_derivative_matrix(nGeo, xGL)
    VdmEqToGL = calc_vandermonde(nGeo, nGeo, wBaryEq, xEq, xGL)
    wbaryGL   = barycentric_weights(nGeo, xGL)

    D_EqToGL  = np.matmul(DGL, VdmEqToGL)

    # Interpolate derivatives on GL (N) to nGeoRef points
    nGeoRef = 3*(nGeo-1)+1
    xAP     = np.zeros(nGeoRef)
    for i in range(nGeoRef):
        xAP[i] = 2. * float(i)/float(nGeoRef-1) - 1.
    VdmGLtoAP = calc_vandermonde(nGeo, nGeoRef, wbaryGL, xGL, xAP)

    # Map all points to tensor product
    elems = mesh_vars.elems
    nodes = mesh_vars.mesh.points

    # Prepare elements for parallel processing
    jacobian_tasks = []

    for iElem, elem in enumerate(elems):
        # Only consider hexahedrons
        if int(elem['Type']) % 100 != 8:
            continue

        # Get the mapping
        linMap = LINMAP(elem['Type'], order=mesh_vars.nGeo)
        mapLin = {k: v for v, k in enumerate(linMap)}

        # Fill the NodeCoords
        nodeCoords = np.zeros((nGeo ** 3, 3), dtype=np.float64)
        for iNode, nodeID in enumerate(elem['Nodes']):
            nodeCoords[mapLin[iNode], :] = nodes[nodeID]

        xGeo = np.zeros((3, nGeo, nGeo, nGeo))
        iNode = 0
        for k in range(nGeo):
            for j in range(nGeo):
                for i in range(nGeo):
                    xGeo[:, i, j, k] = nodeCoords[iNode, :]
                    iNode += 1

        if np_mtp > 0:
            # Add tasks for parallel processing
            jacobian_tasks.append((xGeo, nGeoRef, VdmGLtoAP, D_EqToGL))
        else:
            jac = evaluate_jacobian(xGeo, VdmGLtoAP, D_EqToGL)
            maxJac =  np.max(np.abs(jac))
            minJac =  np.min(       jac)
            jacobian_tasks.append(minJac / maxJac)

    if np_mtp > 0:
        # Run in parallel with a chunk size
        jacs = run_in_parallel(jacobian_tasks, chunk_size=10)
    else:
        jacs = np.array(jacobian_tasks)

    # Plot the histogram of the Jacobians
    plot_histogram(np.array(jacs))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# import os
import subprocess
import sys
import time
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def main() -> None:
    """ Main routine of PyHOPE
    """
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    from pyhope.common.common import DefineCommon, InitCommon
    from pyhope.common.common_vars import Common
    from pyhope.basis.basis_jacobian import CheckJacobians
    from pyhope.io.io import IO, DefineIO, InitIO
    from pyhope.mesh.mesh import DefineMesh, InitMesh, GenerateMesh, RegenerateMesh
    from pyhope.mesh.mesh_connect import ConnectMesh
    from pyhope.mesh.mesh_sides import GenerateSides
    from pyhope.mesh.mesh_sort import SortMesh
    from pyhope.readintools.commandline import CommandLine
    from pyhope.readintools.readintools import DefineConfig, ReadConfig
    # ------------------------------------------------------

    tStart  = time.time()
    process = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], shell=False, stdout=subprocess.PIPE)

    common  = Common()
    program = common.program
    version = common.version
    commit  = process.communicate()[0].strip().decode('ascii')

    with DefineConfig() as dc:
        config.prms = dc
        DefineCommon()
        DefineIO()
        DefineMesh()

    # Parse the command line arguments
    with CommandLine(sys.argv, program, version, commit) as command:
        args = command[0]
        argv = command[1]

    # Exit with version if requested
    if args.version:
        print('{} version {} [commit {}]'.format(program, version, commit))
        sys.exit(0)

    # Check if there are unrecognized arguments
    if len(argv) >= 1:
        print('{} expects exactly one parameter file! Exiting ...'
              .format(program))
        sys.exit()

    with ReadConfig(args.parameter) as rc:
        config.params = rc

    # Print banner
    hopout.header(program, version, commit)

    # Read-in required parameters
    InitCommon()
    InitIO()
    InitMesh()

    # Generate the actual mesh
    GenerateMesh()
    SortMesh()
    GenerateSides()
    RegenerateMesh()
    ConnectMesh()

    # Check the mapping
    CheckJacobians()

    # Output the mesh
    IO()

    tEnd = time.time()
    hopout.end(program, tEnd - tStart)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
import os
import subprocess
import sys
import traceback
from typing import Union
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
from collections import OrderedDict
from configparser import ConfigParser
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


class MultiOrderedDict(OrderedDict):
    """ Add option to repeat the same key multiple times

        Standard ConfigParser only supports one value per key,
        thus overload the ConfigParser with this new dict_type
    """
    def __setitem__(self, key, value):
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            super().__setitem__(key, value)


def strtobool(val: Union[int, bool, str]) -> bool:  # From distutils.util.strtobool() [Python 3.11.2]
    """ Convert a string representation of truth to True or False.
        True values  are 'y', 'yes', 't', 'true', 'on', and '1';
        False values are 'n', 'no' , 'f', 'false', 'off', and '0'.
        Raises ValueError if 'val' is anything else.
    """
    if type(val) is bool:
        return val
    if type(val) is int:
        val = str(val)
    if type(val) is str:
        val = val.lower()

    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError('invalid truth value %r' % (val,))


# ==================================================================================================================================
class DefineConfig:
    """ Provide routines to define all HOPR parameters
    """
    def __init__(self) -> None:
        # Create an empty config dictionary
        self.dict = dict()
        return None

    def __enter__(self) -> dict:
        return self.dict

    def __exit__(self, *args: object) -> None:
        return None


# ==================================================================================================================================
def CheckDefined(name: str, multiple: bool = False, init: bool = False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    # Check if already defined which is allowed if
    # - we are not in init
    # - multiple parameter
    if init:
        if name in config.prms and not multiple:
            hopout.warning('Parameter "{}" already define and not a multiple option, exiting...'.format(name))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)
    else:
        if name not in config.prms:
            hopout.warning('Parameter "{}" is not defined, exiting...'.format(name))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


def CheckUsed(name: str) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if config.prms[name]['counter'] > 1 and not config.prms[name]['multiple']:
        hopout.warning('Parameter "{}" already used and not a multiple option, exiting...'.format(name))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)


def CheckType(name: str, calltype: str) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if config.prms[name]['type'] is not calltype:
        hopout.warning('Call type of parameter "{}" does not match definition, exiting...'.format(name))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)


def CheckDimension(name: str, result: int) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if config.prms[name]['number'] != result:
        hopout.warning('Parameter "{}" has array length mismatch, exiting...'.format(name))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)


def CreateSection(string: str) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple=False, init=True)
    config.prms[string] = dict(type='section', name=string)


def CreateStr(string: str, help: Union[str, None] = None, default: Union[str, None] = None, multiple: bool = False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='str',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateInt(string: str, help: Union[str, None] = None, default: Union[int, None] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='int',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateLogical(string: str, help: Union[str, None] = None, default: Union[bool, None] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='bool',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateIntFromString(string: str, help: Union[str, None] = None, default: Union[str, None] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='int2str',
                               name=string,
                               mapping=dict(),
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateIntOption(string: str, name, number) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    multiple = config.prms[string]['multiple']
    CheckDefined(string, multiple=multiple, init=False)
    config.prms[string]['mapping'].update({number: name})


def CreateRealArray(string: str, nReals, help: Union[str, None] = None, default: Union[str, None] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='realarray',
                               number=nReals,
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateIntArray(string: str, nInts, help: Union[str, None] = None, default: Union[str, None] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='intarray',
                               number=nInts,
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


# ==================================================================================================================================
def CountOption(string: str) -> int:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    from configparser import NoOptionError
    # ------------------------------------------------------

    CheckDefined(string)

    try:
        counter = len([s for s in config.params.get('general', string).split('\n') if s != ''])
    except NoOptionError:
        counter = 0
    return counter


def GetParam(name: str, calltype: str, default: Union[str, None] = None, number: Union[int, None] = None):
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    CheckDefined(name)
    config.prms[name]['counter'] += 1
    CheckUsed(name)
    CheckType(name, calltype)

    if config.params.has_option('general', name):
        if config.prms[name]['multiple']:
            # We can request specific indices
            if number is None: num = config.prms[name]['counter']-1  # noqa: E701
            else:              num = number                          # noqa: E701

            value = [s for s in config.params.get('general', name).split('\n') if s != ''][num]
        else:
            value = config.params.get('general', name)

        # int2str has custom output
        if calltype != 'int2str':
            if calltype == 'bool':
                hopout.printoption(name, '{0:}'.format(value), '*CUSTOM')
            else:
                hopout.printoption(name, value               , '*CUSTOM')
    else:
        if default:
            value = default
        else:
            if config.prms[name]['default'] is not None:
                value = config.prms[name]['default']

                # int2str has custom output
                if calltype != 'int2str':
                    if calltype == 'bool':
                        hopout.printoption(name, '{0:}'.format(value), 'DEFAULT')
                    else:
                        hopout.printoption(name, value               , 'DEFAULT')
            else:
                hopout.warning('Keyword "{}" not found in file and no default given, exiting...'
                               .format(name))
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)
    return value


def GetStr(name: str, default: Union[str, None] = None, number: Union[int, None] = None) -> str:
    value = GetParam(name=name, default=default, number=number, calltype='str')
    return value


def GetInt(name: str, default: Union[str, None] = None, number: Union[int, None] = None) -> int:
    value = GetParam(name=name, default=default, number=number, calltype='int')
    return int(value)


def GetLogical(name: str, default: Union[str, None] = None, number: Union[int, None] = None) -> bool:
    value = GetParam(name=name, default=default, number=number, calltype='bool')
    return strtobool(value)


def GetIntFromStr(name: str, default: Union[str, None] = None, number: Union[int, None] = None) -> int:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    value = GetParam(name=name, default=default, number=number, calltype='int2str')
    # Check if we already received the int. Otherwise, get the value from the
    #  mapping
    mapping = config.prms[name]['mapping']
    if type(value) is int:
        value = value
        hopout.printoption(name, '{} [{}]'.format(value, mapping[value]), 'DEFAULT')
    else:
        if not value.isdigit():
            value = [s for s, v in mapping.items() if v.lower() == value.lower()]
            if len(value) == 0:
                hopout.warning('Unknown value for parameter {}, exiting...'.format(name))
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)
            else:
                value = int(value[0])
                hopout.printoption(name, '{} [{}]'.format(value, mapping[value]), '*CUSTOM')

    return int(value)


def GetRealArray(name: str, default: Union[str, None] = None, number: Union[int, None] = None) -> np.ndarray:
    value = GetParam(name=name, default=default, number=number, calltype='realarray')

    # Split the array definitiosn
    value = value.split('(/')[1]
    value = value.split('/)')[0]

    # Commas separate 1st dimension, double commas separate 2nd dimension
    if ',,' in value:
        value = [s.split(',') for s in value.split(',,')]
        value = np.array(value).astype(float)
    else:
        value = value.split(',')
        value = np.array(value).astype(float)
    CheckDimension(name, value.size)
    return value


def GetIntArray(name: str, default: Union[str, None] = None, number: Union[int, None] = None) -> np.ndarray:
    value = GetParam(name=name, default=default, number=number, calltype='intarray')

    # Split the array definitiosn
    value = value.split('(/')[1]
    value = value.split('/)')[0]

    # Commas separate 1st dimension, double commas separate 2nd dimension
    value = [s.split(',') for s in value.split(',,')]
    value = np.array(value).astype(int)
    # Reduce dimensions
    value = np.concatenate(value).ravel()
    CheckDimension(name, value.size)
    return value


# ==================================================================================================================================
class ReadConfig():
    """ Read an HOPR parameter file

        This file is meant to remain compatible to the HOPR parameter file
        format, so we need some hacks around the INI file format
    """

    def __init__(self, parameter: str) -> None:
        self.parameter = parameter
        return None

    def __enter__(self) -> ConfigParser:
        # Local imports ----------------------------------------
        from pyhope.common.common_vars import Common
        import pyhope.config.config as config
        import pyhope.output.output as hopout
        # ------------------------------------------------------

        parser = ConfigParser(strict=False,
                              comment_prefixes=('#', ';', '!'),
                              inline_comment_prefixes=('#', ';', '!'),
                              dict_type=MultiOrderedDict
                              )

        # Check if the file exists
        if not self.parameter:
            process = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], shell=False, stdout=subprocess.PIPE)
            common  = Common()
            program = common.program
            version = common.version
            commit  = process.communicate()[0].strip().decode('ascii')

            hopout.header(program, version, commit)
            hopout.warning('No parameter file given')
            sys.exit(1)
        if not os.path.isfile(self.parameter):
            process = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], shell=False, stdout=subprocess.PIPE)
            common  = Common()
            program = common.program
            version = common.version
            commit  = process.communicate()[0].strip().decode('ascii')

            hopout.header(program, version, commit)
            hopout.warning('Parameter file [󰇘]/{} does not exist'.format(os.path.basename(self.parameter)))
            sys.exit(1)

        # HOPR does not use conventional sections, so prepend a fake section header
        with open(self.parameter) as stream:
            parser.read_string('[general]\n' + stream.read())

        config.std_length = max(len(s) for s in config.prms.keys())
        config.std_length = max(32, config.std_length+1)

        # Loop over all objects and check if they are provided
        # for key, value in config.prms.items():
        #     if value['type'] == 'section':
        #         hopout.separator()
        #         hopout.info(key)
        #         hopout.separator()
        #         continue
        #
        #     # Check if the key is given in the parameter file
        #     if parser.has_option('general', key):
        #         # Check if the value can be converted
        #         match value['type']:
        #             case 'int':
        #                 try:
        #                     str_int = int(parser.get('general', key))
        #                 except ValueError:
        #                     hopout.warning('Keywords {} cannot be converted to integer'.format(key))
        #
        #         hopout.printoption(key, parser.get('general', key),
        #                            '*CUSTOM', std_length)
        #     # Check if a default option is given
        #     else:
        #         if value['default']:
        #             hopout.printoption(key, value['default'],
        #                                'DEFAULT', std_length)
        #         else:
        #             hopout.warning('Keyword "{}" not found in file, exiting...'
        #                            .format(key))

        return parser

    def __exit__(self, *args: object) -> None:
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
import platform
import subprocess
import sys
from importlib import metadata
from io import TextIOWrapper
from packaging.version import Version
from typing import Union, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def DefineCommon() -> None:
    """ Define general options for the entire program
    """
    # Local imports ----------------------------------------
    from pyhope.readintools.readintools import CreateInt, CreateSection
    # ------------------------------------------------------
    CreateSection('Common')
    CreateInt(      'nThreads',        default=4,     help='Number of threads for multiprocessing')


def InitCommon() -> None:
    """ Readin general option for the entire program
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.common.common_vars as common_vars
    from pyhope.readintools.readintools import GetInt
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('INIT PROGRAM...')

    # Check the number of available threads
    np_mtp = 4
    np_req = GetInt('nThreads')
    match np_req:
        case -1 | 0:  # All available cores / no multiprocessing
            np_mtp = np_req
        case _:       # Check if the number of requested processes can be provided
            # os.affinity is Linux only
            try:
                np_aff = len(os.sched_getaffinity(0))
            except AttributeError:
                np_aff = os.cpu_count() or 1
            np_mtp = min(np_req, np_aff)

    # If running under debugger, multiprocessing is not available
    if DebugEnabled():
        hopout.info('Debugger detected, disabling multiprocessing!')
        np_mtp = 0

    # Actually overwrite the global value
    common_vars.np_mtp = np_mtp

    # Check if we are using the NRG Gmsh version and install it if not
    PkgsCheckGmsh()

    hopout.info('INIT PROGRAM DONE!')


def DebugEnabled() -> bool:
    """ Check if program runs with debugger attached
        > https://stackoverflow.com/a/77627075/23851165
    """
    try:
        if sys.gettrace() is not None:
            return True
    except AttributeError:
        pass

    try:
        if sys.monitoring.get_tool(sys.monitoring.DEBUGGER_ID) is not None:
            return True
    except AttributeError:
        pass

    return False


def IsInteractive():
    return cast(TextIOWrapper, sys.__stdin__).isatty()


def PkgsMetaData(pkgs, classifier) -> Union[bool, None]:
    """ Check if the package contains a given classifier
    """
    try:
        meta = metadata.metadata(pkgs)
        classifiers = meta.get_all('Classifier', [])
        return classifier in classifiers

    except metadata.PackageNotFoundError:
        return None


def PkgsMetaVersion(pkgs) -> Union[str, None]:
    """ Check the package version
    """
    try:
        version = metadata.version(pkgs)
        return version

    except metadata.PackageNotFoundError:
        return None


def PkgsCheckGmsh() -> None:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    gmsh_version  = PkgsMetaVersion('gmsh')
    if gmsh_version is None:
        # Gmsh is not installed
        if IsInteractive():
            warning = 'Gmsh is not installed. For compatibility, the NRG Gmsh version will be installed. Continue? (Y/n):'
            response = input('\n' + hopout.warn(warning) + '\n')
            if response.lower() in ['yes', 'y', '']:
                PkgsInstallGmsh()
                return None
        else:
            hopout.warning('Gmsh is not installed, exiting...')
            sys.exit(1)

    gmsh_version = cast(str, gmsh_version)
    # Assume that newer versions have updated CGNS
    gmsh_expected = '4.14'
    if Version(gmsh_version) > Version(gmsh_expected):
        return None

    # Check if the installed version is the NRG version
    if PkgsMetaData('gmsh', 'Intended Audience: NRG'):
        return None

    # Check the current platform
    system = platform.system()
    arch   = platform.machine()

    if system != 'Linux' or arch != 'x86_64':
        warning = hopout.warn(f'Detected non-NRG Gmsh version on unsupported platform [{system}/{arch}]. ' +
                              'Functionality may be limited.')
        print(warning)
        return None

    if not PkgsMetaData('gmsh', 'Intended Audience: NRG'):
        if IsInteractive():
            warning  = 'Detected Gmsh package uses an outdated CGNS (v3.4). For compatibility, ' \
                       'the package will be uninstalled and replaced with the updated NRG GMSH ' \
                       'version. Continue? (Y/n):'
            response = input('\n' + hopout.warn(warning) + '\n')
            if response.lower() in ['yes', 'y', '']:
                PkgsInstallGmsh()
                return None
        else:
            warning = hopout.warn('Detected Gmsh package uses an outdated CGNS (v3.4). Functionality may be limited.')
            print(warning)
            return None


def PkgsInstallGmsh():
    # Local imports ----------------------------------------
    import hashlib
    import tempfile
    import pyhope.output.output as hopout
    from pyhope.common.common_vars import Gitlab
    # ------------------------------------------------------

    # Gitlab "python-gmsh" access
    lfs = 'yes'
    lib = 'gmsh-{}-py3-none-linux_x86_64.whl'.format(Gitlab.LIB_VERSION)

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as path:
        pkgs = os.path.join(path, lib)
        curl = [f'curl https://{Gitlab.LIB_GITLAB}/api/v4/projects/{Gitlab.LIB_PROJECT}/repository/files/{lib}/raw?lfs={lfs} --output {pkgs}']  # noqa: E501
        subprocess.run(curl, check=True, shell=True)

        # Compare the hash
        # > Initialize a new sha256 hash
        sha256 = hashlib.sha256()
        with open(pkgs, 'rb') as f:
            # Read and update hash string value in blocks of 4K
            for chunk in iter(lambda: f.read(4096), b""):
                sha256.update(chunk)

        if sha256.hexdigest() == Gitlab.LIB_HASH:
            hopout.info('Hash matches, installing Gmsh wheel...')
        else:
            hopout.warning('Hash mismatch, exiting...')
            sys.exit(1)

        # Remove the old version
        try:
            meta = metadata.metadata('gmsh')
            if meta is not None:
                subprocess.run([sys.executable, '-m', 'pip', 'uninstall', '-y', 'gmsh'], check=True)

        except metadata.PackageNotFoundError:
            pass

        # Install the package in the current environment
        subprocess.run([sys.executable, '-m', 'pip', 'install', pkgs], check=True)


# > https://stackoverflow.com/a/5419576/23851165
# def object_meth(object) -> list:
#     methods = [method_name for method_name in dir(object)
#                if '__' not in method_name]
#     return methods


def find_key(dict: dict, item) -> int | None:
    """ Find the first occurrence of a key in dictionary
    """
    if type(item) is np.ndarray:
        for key, val in dict.items():
            if np.all(val == item):
                return key
    else:
        for key, val in dict.items():
            if        val == item :  # noqa: E271
                return key
    return None


def find_keys(dict: dict, item) -> list | None:
    """ Find all occurrence of a key in dictionary
    """
    if type(item) is np.ndarray:
        keys = [key for key, val in dict.items() if np.all(val == item)]
        if len(keys) > 0:
            return keys
    else:
        keys = [key for key, val in dict.items() if        val == item ]  # noqa: E271
        if len(keys) > 0:
            return keys
    return None


# def find_value(dict, item):
#     """ Find key by value in dictionary
#     """
#     return dict.keys()[dict.values().index(item)]


def find_index(seq, item) -> int:
    """ Find the first occurrences of a key in a list
    """
    if type(seq) is np.ndarray:
        seq = seq.tolist()

    if type(item) is np.ndarray:
        for index, val in enumerate(seq):
            if np.all(val == item):
                return index
    else:
        for index, val in enumerate(seq):
            if        val == item :  # noqa: E271
                return index
    return -1


def find_indices(seq, item) -> list:
    """ Find all occurrences of a key in a list
    """
    if type(seq) is np.ndarray:
        seq = seq.tolist()

    start_at = -1
    locs = []
    while True:
        try:
            loc = seq.index(item, start_at+1)
        except ValueError:
            break
        else:
            locs.append(loc)
            start_at = loc
    return locs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def ElemTypes(num: int) -> dict[str, str]:
    types = [{ 'ElemTypeCGNS': 'ElementTypeNull', 'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 0
             { 'ElemTypeCGNS': 'ElementTypeUser', 'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 1
             { 'ElemTypeCGNS': 'NODE',            'ElemTypeMeshIO': 'vertex',       'Nodes': 1  },  # 2
             { 'ElemTypeCGNS': 'BAR_2',           'ElemTypeMeshIO': 'line',         'Nodes': 2  },  # 3
             { 'ElemTypeCGNS': 'BAR_3',           'ElemTypeMeshIO': 'line3',        'Nodes': 3  },  # 4
             { 'ElemTypeCGNS': 'TRI_3',           'ElemTypeMeshIO': 'triangle',     'Nodes': 3  },  # 5
             { 'ElemTypeCGNS': 'TRI_6',           'ElemTypeMeshIO': 'triangle6',    'Nodes': 6  },  # 6
             { 'ElemTypeCGNS': 'QUAD_4',          'ElemTypeMeshIO': 'quad',         'Nodes': 4  },  # 7
             { 'ElemTypeCGNS': 'QUAD_8',          'ElemTypeMeshIO': 'quad8',        'Nodes': 8  },  # 8
             { 'ElemTypeCGNS': 'QUAD_9',          'ElemTypeMeshIO': 'quad9',        'Nodes': 9  },  # 9
             { 'ElemTypeCGNS': 'TETRA_4',         'ElemTypeMeshIO': 'tetra',        'Nodes': 4  },  # 10
             { 'ElemTypeCGNS': 'TETRA_10',        'ElemTypeMeshIO': 'tetra10',      'Nodes': 10 },  # 11
             { 'ElemTypeCGNS': 'PYRA_5',          'ElemTypeMeshIO': 'pyramid',      'Nodes': 5  },  # 12
             { 'ElemTypeCGNS': 'PYRA_14',         'ElemTypeMeshIO': 'pyramid14',    'Nodes': 14 },  # 13
             { 'ElemTypeCGNS': 'PENTA_6',         'ElemTypeMeshIO': 'wedge',        'Nodes': 6  },  # 14
             { 'ElemTypeCGNS': 'PENTA_15',        'ElemTypeMeshIO': 'wedge15',      'Nodes': 15 },  # 15
             { 'ElemTypeCGNS': 'PENTA_18',        'ElemTypeMeshIO': 'wedge18',      'Nodes': 18 },  # 16
             { 'ElemTypeCGNS': 'HEXA_8',          'ElemTypeMeshIO': 'hexahedron',   'Nodes': 8  },  # 17
             { 'ElemTypeCGNS': 'HEXA_20',         'ElemTypeMeshIO': 'hexahedron20', 'Nodes': 20 },  # 18
             { 'ElemTypeCGNS': 'HEXA_27',         'ElemTypeMeshIO': 'hexahedron27', 'Nodes': 27 },  # 19
             { 'ElemTypeCGNS': 'MIXED',           'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 20
             { 'ElemTypeCGNS': 'PYRA_13',         'ElemTypeMeshIO': 'pyramid13',    'Nodes': 13 },  # 21
             { 'ElemTypeCGNS': 'NGON_n',          'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 22
             { 'ElemTypeCGNS': 'NFACE_n',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 23
             { 'ElemTypeCGNS': 'BAR_4',           'ElemTypeMeshIO': 'line4',        'Nodes': 4  },  # 24
             { 'ElemTypeCGNS': 'TRI_9',           'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 25
             { 'ElemTypeCGNS': 'TRI_10',          'ElemTypeMeshIO': 'triangle10',   'Nodes': 10 },  # 26
             { 'ElemTypeCGNS': 'QUAD_12',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 27
             { 'ElemTypeCGNS': 'QUAD_16',         'ElemTypeMeshIO': 'quad16',       'Nodes': 16 },  # 28
             { 'ElemTypeCGNS': 'TETRA_16',        'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 29
             { 'ElemTypeCGNS': 'TETRA_20',        'ElemTypeMeshIO': 'tetra20',      'Nodes': 20 },  # 30
             { 'ElemTypeCGNS': 'PYRA_21',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 31
             { 'ElemTypeCGNS': 'PYRA_29',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 32
             { 'ElemTypeCGNS': 'PYRA_30',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 33
             { 'ElemTypeCGNS': 'PENTA_24',        'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 34
             { 'ElemTypeCGNS': 'PENTA_38',        'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 35
             { 'ElemTypeCGNS': 'PENTA_40',        'ElemTypeMeshIO': 'wedge40',      'Nodes': 40 },  # 36
             { 'ElemTypeCGNS': 'HEXA_32',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 37
             { 'ElemTypeCGNS': 'HEXA_56',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 38
             { 'ElemTypeCGNS': 'HEXA_64',         'ElemTypeMeshIO': 'hexahedron64', 'Nodes': 64 },  # 39
             { 'ElemTypeCGNS': 'Null',            'ElemTypeMeshIO': 'hexahedron24', 'Nodes': 24 },  # 40
             ]
    return types[num]


def spiral_matrix(n: int) -> np.ndarray:
    """ Print a spiral matrix of order n
        > https://rosettacode.org/wiki/Spiral_matrix#Simple_solution
    """
    m = [[0] * n for _ in range(n)]
    dx, dy  = [0, 1, 0, -1], [1, 0, -1, 0]
    x, y, c =  0, -1, 1
    for i in range(n + n - 1):
        for j in range((n + n - i) // 2):
            x += dx[i % 4]
            y += dy[i % 4]
            m[x][y] = c
            c += 1
    return np.array(m)


def edgePointCGNS(order: int, edge: int, node: int) -> np.ndarray:
    match edge:
        case 0:  # z- / base
            return np.array([node      , 0         ], dtype=int)
        case 1:  # y+ / base
            return np.array([order     , node      ], dtype=int)
        case 2:  # z+ / base
            return np.array([order-node, order     ], dtype=int)
        case 3:  # y- / base
            return np.array([0         , order-node], dtype=int)
        case _:
            sys.exit(1)


def genHEXMAPCGNS(order: int) -> None:
    """ CGNS -> IJK ordering for high-order hexahedrons
        > Losely based on [Gmsh] "generatePointsHexCGNS"
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    map = np.zeros((order, order, order), dtype=int)

    if order == 1:
        map[0, 0, 0] = 0
        mesh_vars.HEXMAP = map
        return None

    # Principal vertices
    map[0      , 0      , 0      ] = 1
    map[order-1, 0      , 0      ] = 2
    map[order-1, order-1, 0      ] = 3
    map[0      , order-1, 0      ] = 4
    map[0      , 0      , order-1] = 5
    map[order-1, 0      , order-1] = 6
    map[order-1, order-1, order-1] = 7
    map[0      , order-1, order-1] = 8

    if order == 2:
        # Python indexing, 1 -> 0
        map -= 1
        # Reshape into 1D array, tensor-product style
        tensor = []
        for k in range(order):
            for j in range(order):
                for i in range(order):
                    tensor.append(int(map[i, j, k]))

        mesh_vars.HEXMAP = tensor
        return None

    # Internal points of base quadrangle edges (x-)
    count = 8
    for iFace in range(4):
        for iNode in range(1, order-1):
            # Assemble mapping to tuple, base quadrangle -> z = 0
            count += 1
            edge  = edgePointCGNS(order-1, iFace, iNode)
            index = (int(edge[0]), int(edge[1]), 0)
            map[index] = count

    # Internal points of mounting edges
    for iFace in range(4):
        edge  = edgePointCGNS(order-1, (iFace+3) % 4, order-1)
        for iNode in range(1, order-1):
            # Assemble mapping to tuple, mounting edges -> z ascending
            count += 1
            index = (int(edge[0]), int(edge[1]), iNode)
            map[index] = count

    # Internal points of top quadrangle edges
    for iFace in range(4):
        for iNode in range(1, order-1):
            # Assemble mapping to tuple, top  quadrangle -> z = order
            count += 1
            edge  = edgePointCGNS(order-1, iFace, iNode)
            index = (int(edge[0]), int(edge[1]), order-1)
            map[index] = count

    # Internal points of base quadrangle
    k    = 0
    # Fill in spirals
    face = spiral_matrix(order-2)
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            index = (i+1  , j+1  , k    )
            map[index] = count + face[j, i]
    count += (order-2)**2

    # Internal points of upstanding faces
    # > y- face
    k    = 0
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            index = (i+1  , k    , j+1  )
            map[index] = count + face[j, i]
    count += (order-2)**2

    # > x+ face
    k    = order-1
    for j in range(order-2):
        for i in range(order-2):
            index = (k    , i+1  , j+1  )
            map[index] = count + face[j, i]
    count += (order-2)**2

    # > y+ face
    k      = order-1
    face_r = np.rot90(face, k=2, axes=(1, 0))
    for j in range(order-2):
        for i in range(order-2):
            index = (i+1  , k    , j+1  )
            map[index] = count + face_r[order-3-j, i]
    count += (order-2)**2

    # > x- face
    k      = 0
    for j in range(order-2):
        for i in range(order-2):
            index = (k    , i+1  , j+1  )
            map[index] = count + face_r[order-3-j, i]
    count += (order-2)**2

    # Internal points of top  quadrangle
    k    = order-1
    # Fill in spirals
    face = spiral_matrix(order-2)
    # Fill the map
    for j in range(order-2):
        for i in range(order-2):
            index = (i+1  , j+1  , k    )
            map[index] = count + face[j, i]
    count += (order-2)**2

    # Internal volume points as a succession of internal planes
    for k in range(1, order-1):
        for j in range(order-2):
            for i in range(order-2):
                index = (i+1  , j+1  , k    )
                map[index] = count + face[j, i]
        count += (order-2)**2

    # Python indexing, 1 -> 0
    map -= 1

    # Reshape into 1D array, tensor-product style
    tensor = []
    for k in range(order):
        for j in range(order):
            for i in range(order):
                tensor.append(int(map[i, j, k]))
    mesh_vars.HEXMAP = tensor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import copy
import math
import sys
import traceback
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import gmsh
import meshio
import pygmsh
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def MeshCartesian() -> meshio._mesh.Mesh:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common import find_index, find_indices
    from pyhope.io.io_vars import debugvisu
    from pyhope.mesh.mesh_common import edge_to_dir, face_to_corner, face_to_edge, faces
    from pyhope.readintools.readintools import CountOption, GetInt, GetIntArray, GetRealArray, GetStr
    # ------------------------------------------------------

    gmsh.initialize()
    if not debugvisu:
        # Hide the GMSH debug output
        gmsh.option.setNumber('General.Terminal', 0)

    hopout.sep()

    nZones = GetInt('nZones')

    offsetp  = 0
    offsets  = 0

    # GMSH only supports mesh elements within a single model
    # > https://gitlab.onelab.info/gmsh/gmsh/-/issues/2836
    gmsh.model.add('Domain')
    gmsh.model.set_current('Domain')
    bcZones = [list() for _ in range(nZones)]

    for zone in range(nZones):
        hopout.routine('Generating zone {}'.format(zone+1))

        corners = GetRealArray('Corner', number=zone)
        nElems  = GetIntArray( 'nElems', number=zone)

        # Create all the corner points
        p = [None for _ in range(len(corners))]
        for index, corner in enumerate(corners):
            p[index] = gmsh.model.geo.addPoint(*corner, tag=offsetp+index+1)

        # Connect the corner points
        e = [None for _ in range(12)]
        # First, the plane surface
        for i in range(2):
            for j in range(4):
                e[j + i*4] = gmsh.model.geo.addLine(p[j + i*4], p[(j+1) % 4 + i*4])
        # Then, the connection
        for j in range(4):
            e[j+8] = gmsh.model.geo.addLine(p[j], p[j+4])

        # We need to define the curves as transfinite curves
        # and set the correct spacing from the parameter file
        for index, line in enumerate(e):
            # We set the number of nodes, so Elems+1
            gmsh.model.geo.mesh.setTransfiniteCurve(line, nElems[edge_to_dir(index)]+1)

        # Create the curve loop
        el = [None for _ in range(len(faces()))]
        for index, face in enumerate(faces()):
            el[index] = gmsh.model.geo.addCurveLoop([math.copysign(e[abs(s)], s) for s in face_to_edge(face)])

        # Create the surfaces
        s = [None for _ in range(len(faces()))]
        for index, surface in enumerate(s):
            s[index] = gmsh.model.geo.addPlaneSurface([el[index]], tag=offsets+index+1)

        # We need to define the surfaces as transfinite surface
        for index, face in enumerate(faces()):
            gmsh.model.geo.mesh.setTransfiniteSurface(offsets+index+1, face, [p[s] for s in face_to_corner(face)])
            gmsh.model.geo.mesh.setRecombine(2, 1)

        # Create the surface loop
        gmsh.model.geo.addSurfaceLoop([s for s in s], zone+1)

        gmsh.model.geo.synchronize()

        # Create the volume
        gmsh.model.geo.addVolume([zone+1], zone+1)

        # We need to define the volume as transfinite volume
        gmsh.model.geo.mesh.setTransfiniteVolume(zone+1)
        gmsh.model.geo.mesh.setRecombine(3, 1)

        # Calculate all offsets
        offsetp += len(corners)
        offsets += len(faces())

        # Read the BCs for the zone
        # > Need to wait with defining phyiscal boundaries until all zones are created
        bcZones[zone] = [int(s) for s in GetIntArray('BCIndex')]

    # At this point, we can create a "Physical Group" corresponding
    # to the boundaries. This requires a synchronize call!
    gmsh.model.geo.synchronize()

    hopout.sep()
    hopout.routine('Setting boundary conditions')
    hopout.sep()
    nBCs = CountOption('BoundaryName')
    mesh_vars.bcs = [dict() for _ in range(nBCs)]
    bcs = mesh_vars.bcs

    for iBC, bc in enumerate(bcs):
        bcs[iBC]['Name'] = GetStr('BoundaryName', number=iBC)
        bcs[iBC]['BCID'] = iBC + 1
        bcs[iBC]['Type'] = GetIntArray('BoundaryType', number=iBC)

    nVVs = CountOption('vv')
    mesh_vars.vvs = [dict() for _ in range(nVVs)]
    vvs = mesh_vars.vvs
    for iVV, vv in enumerate(vvs):
        vvs[iVV] = dict()
        vvs[iVV]['Dir'] = GetRealArray('vv', number=iVV)
    if len(vvs) > 0:
        hopout.sep()

    # Flatten the BC array, the surface numbering follows from the 2-D ordering
    bcIndex = [item for row in bcZones for item in row]

    bc = [None for _ in range(max(bcIndex))]
    for iBC in range(max(bcIndex)):
        # if mesh_vars.bcs[iBC-1] is None:
        if 'Name' not in bcs[iBC]:
            continue

        # Format [dim of group, list, name)
        # > Here, we return ALL surfaces on the BC, irrespective of the zone
        surfID  = [s+1 for s in find_indices(bcIndex, iBC+1)]
        bc[iBC] = gmsh.model.addPhysicalGroup(2, surfID, name=bcs[iBC]['Name'])

        # For periodic sides, we need to impose the periodicity constraint
        if bcs[iBC]['Type'][0] == 1:
            # > Periodicity transform is provided as a 4x4 affine transformation matrix, given by row
            # > Rotation matrix [columns 0-2], translation vector [column 3], bottom row [0, 0, 0, 1]

            # Only define the positive translation
            if bcs[iBC]['Type'][3] > 0:
                pass
            elif bcs[iBC]['Type'][3] == 0:
                hopout.warning('BC "{}" has no periodic vector given, exiting...'.format(iBC + 1))
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)
            else:
                continue

            hopout.routine('Generated periodicity constraint with vector {}'.format(vvs[int(bcs[iBC]['Type'][3])-1]['Dir']))

            translation = [1., 0., 0., float(vvs[int(bcs[iBC]['Type'][3])-1]['Dir'][0]),
                           0., 1., 0., float(vvs[int(bcs[iBC]['Type'][3])-1]['Dir'][1]),
                           0., 0., 1., float(vvs[int(bcs[iBC]['Type'][3])-1]['Dir'][2]),
                           0., 0., 0., 1.]

            # Find the opposing side(s)
            # > copy, otherwise we modify bcs
            nbType     = copy.copy(bcs[iBC]['Type'])
            nbType[3] *= -1
            nbBCID     = find_index([s['Type'] for s in bcs], nbType)
            # nbSurfID can hold multiple surfaces, depending on the number of zones
            # > find_indices returns all we need!
            nbSurfID   = [s+1 for s in find_indices(bcIndex, nbBCID+1)]

            # Connect positive to negative side
            gmsh.model.mesh.setPeriodic(2, nbSurfID, surfID, translation)

    # To generate connect the generated cells, we can simply set
    gmsh.option.setNumber('Mesh.RecombineAll', 1)
    # Force Gmsh to output all mesh elements
    gmsh.option.setNumber('Mesh.SaveAll', 1)

    # Set the element order
    # > Technically, this is only required in generate_mesh but let's be precise here
    gmsh.model.mesh.setOrder(mesh_vars.nGeo)

    gmsh.model.geo.synchronize()

    # PyGMSH returns a meshio.mesh datatype
    mesh = pygmsh.geo.Geometry().generate_mesh(order=mesh_vars.nGeo)

    if debugvisu:
        gmsh.fltk.run()

    # Finally done with GMSH, finalize
    gmsh.finalize()

    # Final count
    nElems = 0
    for iType, cellType in enumerate(mesh.cells):
        if any(s in cellType.type for s in mesh_vars.ELEM.type.keys()):
            nElems += mesh.get_cells_type(cellType.type).shape[0]
    hopout.sep()
    hopout.routine('Generated mesh with {} cells'.format(nElems))
    hopout.sep()

    return mesh
